from abc import ABC, abstractmethod
from typing import List, Optional, Tuple

from lsprotocol.types import Range, TextDocumentSaveReason

Uri = str
OldUri = Uri
NewUri = Uri
FileRenameMapping = Tuple[OldUri, NewUri]


class LspClient(ABC):

    @abstractmethod
    def text_document_did_open(
            self,
            document_id: int,
            uri: str,
            language_id: str,
            version: int,
            text: str
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def text_document_did_close(
            self,
            document_id: int,
            uri: Optional[str]
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def text_document_will_save(
            self,
            uri: str,
            reason: TextDocumentSaveReason
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def text_document_will_save_wait_until(
            self,
            uri: str,
            reason: TextDocumentSaveReason
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def text_document_did_save(
            self,
            uri: str,
            text: str
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def text_document_did_change(
            self,
            uri: str,
            version: int,
            start_line: int,
            start_column: int,
            end_line: int,
            end_column: int,
            full_text: str,
            part_text: str
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def workspace_will_create_files(self, files: List[Uri]) -> None:
        raise NotImplementedError

    @abstractmethod
    def workspace_did_create_files(self, files: List[Uri]) -> None:
        raise NotImplementedError

    @abstractmethod
    def workspace_will_rename_files(self, files: List[FileRenameMapping]) -> None:
        raise NotImplementedError

    @abstractmethod
    def workspace_did_rename_files(self, files: List[FileRenameMapping]) -> None:
        raise NotImplementedError

    @abstractmethod
    def workspace_will_delete_files(self, files: List[Uri]) -> None:
        raise NotImplementedError

    @abstractmethod
    def workspace_did_delete_files(self, files: List[Uri]) -> None:
        raise NotImplementedError

    @abstractmethod
    def goto_definition(self, uri: str, line: int, column: int) -> None:
        raise NotImplementedError

    @abstractmethod
    def rename(self, uri: str, line: int, column: int, new_name: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def highlight_symbol(self, uri: str, line: int, column: int) -> None:
        raise NotImplementedError

    @abstractmethod
    def hover(self, uri: str, line: int, column: int) -> None:
        raise NotImplementedError

    @abstractmethod
    def semantic_highlight(self, uri: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def complete(self, uri: str, line: int, column: int) -> None:
        raise NotImplementedError

    @abstractmethod
    def format(self, uri: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def format_range(self, uri: str, selection: Range) -> None:
        raise NotImplementedError



from pathlib import Path

from llanguage_server.cxx.visitors import BaseCPlusPlusLspVisitor
from llanguage_server.lsp.datatypes import LspSpec
from llanguage_server.lsp.utils import (method_to_camel_case, receive_fn,
                                        send_fn)
from llanguage_server.lsp.visitors import LspAnalysisPipeline
from llanguage_server.lsp.datatypes import LspSpec, LspSymbol


class CPlusPlusLspLanguageServerHeaderGenerator(BaseCPlusPlusLspVisitor):

    def __init__(
            self,
            output_dir: Path,
            schema: LspSpec,
            pipeline: LspAnalysisPipeline,
            namespace: str
    ) -> None:
        specification_source = output_dir / "lsp_language_server.h"
        super().__init__(specification_source, schema, pipeline, namespace)

    def generate_receive_request(self, request_spec: LspSpec) -> None:
        request_method = request_spec["method"]
        request_name = method_to_camel_case(request_method)
        result_name = f'{request_name}Result'
        params_spec = request_spec.get("params", None)
        params = [
            'const RequestMessage &request',
        ]
        if params_spec is not None:
            params.append(f'{params_spec["name"]} &params',)
        self.gen_fn_decl(
            receive_fn(request_method),
            result_name,
            params=params,
            docs=request_spec.get("documentation", None),
            virtual=True
        )
        self.newline()

    def generate_send_request(self, request_spec: LspSpec) -> None:
        request_method = request_spec["method"]
        send_request = send_fn(request_method)
        params_spec = request_spec.get("params", None)
        params = []
        if params_spec is not None:
            params.append(f'{params_spec["name"]} &params')
        self.gen_fn_decl(
            send_request,
            'int',
            params=params,
            docs=request_spec.get("documentation", None),
            virtual=True
        )
        self.newline()

    def generate_receive_response(self, request_spec: LspSpec) -> None:
        request_method = request_spec["method"]
        request_name = method_to_camel_case(request_method)
        result_spec = request_spec.get("result", None)
        params = [
            'const RequestMessage &request',
            'const ResponseMessage &response',
        ]
        if result_spec is not None:
            result_name = f'{request_name}Result'
            if result_spec["kind"] == "base":
                params.append(f'{result_name} params')
            else:
                params.append(f'{result_name} &params')
        self.gen_fn_decl(
            receive_fn(request_method),
            params=params,
            virtual=True
        )
        self.newline()

    def generate_outgoing_request_handlers(self, request_spec: LspSpec) -> None:
        self.generate_send_request(request_spec)
        self.generate_receive_response(request_spec)

    def visit_request_symbol(self, request_symbol: LspSymbol) -> None:
        request_spec = request_symbol.spec
        match request_spec["messageDirection"]:
            case "clientToServer":
                self.generate_receive_request(request_spec)
            case "serverToClient":
                self.generate_outgoing_request_handlers(request_spec)
            case "both":
                self.generate_receive_request(request_spec)
                self.generate_outgoing_request_handlers(request_spec)
            case _:
                raise ValueError(
                    f'Unsupported messageDirection: {request_spec["messageDirection"]}'
                )

    def generate_receive_notification(self, notification_spec: LspSpec) -> None:
        notification_method = notification_spec["method"]
        params_spec = notification_spec.get("params", None)
        params = [
            'const NotificationMessage &notification',
        ]
        if params_spec is not None:
            params.append(f'{params_spec["name"]} &params')
        self.gen_fn_decl(
            receive_fn(notification_method),
            params=params,
            docs=notification_spec.get("documentation", None),
            virtual=True
        )
        self.newline()

    def generate_send_notification(self, notification_spec: LspSpec) -> None:
        notification_method = notification_spec["method"]
        params_spec = notification_spec.get("params", None)
        params=[]
        if params_spec is not None:
            params.append(f'{params_spec["name"]} &params')
        self.gen_fn_decl(
            send_fn(notification_method),
            params=params,
            docs=notification_spec.get("documentation", None),
            virtual=True
        )
        self.newline()

    def visit_notification_symbol(self, notification_symbol: LspSymbol) -> None:
        notification_spec = notification_symbol.spec
        match notification_spec["messageDirection"]:
            case "clientToServer":
                self.generate_receive_notification(notification_spec)
            case "serverToClient":
                self.generate_send_notification(notification_spec)
            case "both":
                self.generate_receive_notification(notification_spec)
                self.generate_send_notification(notification_spec)
            case _:
                raise ValueError(
                    f'Unsupported messageDirection: {notification_spec["messageDirection"]}'
                )

    def generate_dispatch_request(self) -> None:
        self.gen_fn_decl('dispatch', params=[
            'ResponseMessage &response',
            'RequestMessage &request',
            'IncomingRequest method',
        ])
        self.newline()

    def generate_dispatch_notification(self) -> None:
        self.gen_fn_decl('dispatch', params=[
            'ResponseMessage &response',
            'NotificationMessage &notification',
            'IncomingNotification method',
        ])
        self.newline()

    def generate_dispatch_response(self) -> None:
        self.gen_fn_decl('dispatch', params=[
            'ResponseMessage &response',
            'const std::string &method',
        ])
        self.newline()

    def generate_require_request_params(self) -> None:
        self.gen_fn_decl(
            'requireMessageParams',
            ret_type='MessageParams &',
            params=[
                'RequestMessage &request',
            ],
            specs='const'
        )
        self.newline()

    def generate_require_notification_params(self) -> None:
        self.gen_fn_decl(
            'requireMessageParams',
            ret_type='MessageParams &',
            params=[
                'NotificationMessage &notification',
            ],
            specs='const'
        )
        self.newline()

    def generate_constructor(self) -> None:
        self.gen_constructor_decl('LspLanguageServer', params=[
            'ls::MessageQueue &incomingMessages',
            'ls::MessageQueue &outgoingMessages',
            'lsl::Logger &logger',
        ])
        self.newline()

    def generate_fields(self) -> None:
        self.write('lsl::Logger logger;')
        self.write('LspTransformer transformer;')
        self.write('LspJsonSerializer serializer;')
        self.write('std::atomic_int serialRequestId = 0;')
        self.write('std::map<int, std::shared_ptr<RequestMessage>> requestsById;')
        self.write('std::shared_mutex requestMutex;')
        self.newline()

    def generate_next_request_id(self) -> None:
        self.gen_fn_decl(
            'nextRequestId',
            ret_type='int',
        )
        self.newline()

    def generate_send_request_message(self) -> None:
        self.gen_fn_decl(
            'send',
            params=[
                'const RequestMessage &request',
            ],
            virtual=True,
        )
        self.newline()

    def generate_build_request(self) -> None:
        self.gen_fn_decl(
            'buildRequest',
            'RequestMessage',
            params=[
                'const std::string &method',
            ],
        )
        self.newline()

    def generate_send_custom_request(self) -> None:
        self.gen_fn_decl(
            'sendRequest',
            'int',
            params=[
                'const std::string &method',
                'MessageParams &params',
            ],
        )
        self.newline()

    def generate_send_custom_request_no_params(self) -> None:
        self.gen_fn_decl(
            'sendRequest',
            'int',
            params=[
                'const std::string &method',
            ],
        )
        self.newline()

    def generate_send_notification_message(self) -> None:
        self.gen_fn_decl(
            'send',
            params=[
                'const NotificationMessage &notification',
            ],
        )
        self.newline()

    def generate_build_notification(self) -> None:
        self.gen_fn_decl(
            'buildNotification',
            'NotificationMessage',
            params=[
                'const std::string &method',
            ],
        )
        self.newline()

    def generate_send_custom_notification(self) -> None:
        self.gen_fn_decl(
            'sendNotification',
            params=[
                'const std::string &method',
                'MessageParams &params',
            ],
        )
        self.newline()

    def generate_send_custom_notification_no_params(self) -> None:
        self.gen_fn_decl(
            'sendNotification',
            params=[
                'const std::string &method',
            ],
        )
        self.newline()

    def generate_code(self) -> None:
        print(f'Generating: {self.file_path}')
        self.generate_disclaimer()
        self.pragma_once()
        self.newline()
        self.gen_include('atomic')
        self.gen_include('map')
        self.gen_include('memory')
        self.gen_include('shared_mutex')
        self.newline()
        self.gen_include('server/language_server.h')
        self.gen_include('server/logger.h')
        self.gen_include('server/lsp_json_serializer.h')
        self.gen_include('server/lsp_specification.h')
        self.gen_include('server/lsp_transformer.h')
        self.newline()
        with self.gen_namespace(self.namespace):
            self.write('namespace ls = LCompilers::LLanguageServer;')
            self.write('namespace lsl = LCompilers::LLanguageServer::Logging;')
            self.newline()
            with self.gen_class('LspLanguageServer', sups=['ls::LanguageServer']):
                with self.gen_protected():
                    self.generate_constructor()
                    self.generate_fields()
                    self.generate_next_request_id()
                    self.generate_send_request_message()
                    self.generate_build_request()
                    self.generate_send_custom_request()
                    self.generate_send_custom_request_no_params()
                    self.generate_send_notification_message()
                    self.generate_build_notification()
                    self.generate_send_custom_notification()
                    self.generate_send_custom_notification_no_params()
                    self.generate_dispatch_request()
                    self.generate_dispatch_notification()
                    self.generate_dispatch_response()
                    self.generate_require_request_params()
                    self.generate_require_notification_params()
                    super().generate_code()



from pathlib import Path
from typing import Set

from llanguage_server.cxx.visitors import BaseCPlusPlusLspVisitor
from llanguage_server.lsp.datatypes import LspSpec, LspSymbol, LspSymbolKind
from llanguage_server.lsp.utils import (lower_first, rename_enum, rename_type,
                                        upper_first)
from llanguage_server.lsp.visitors import LspAnalysisPipeline


class CPlusPlusSpecificationHeaderGenerator(BaseCPlusPlusLspVisitor):

    def __init__(
            self,
            output_dir: Path,
            schema: LspSpec,
            pipeline: LspAnalysisPipeline,
            namespace: str
    ) -> None:
        specification_header = output_dir / "lsp_specification.h"
        super().__init__(specification_header, schema, pipeline, namespace)

    def generate_struct_field(
            self,
            struct_symbol: LspSymbol,
            field_spec: LspSpec
    ) -> None:
        field_name: str = field_spec["name"]
        wrap_ptr = False
        wrap_nested_ptr = False
        match field_spec["type"]["kind"]:
            case "reference":
                dependency = self.resolve(field_spec["type"]["name"])
                wrap_ptr = self.has_cycle(struct_symbol, dependency)
            case "array":
                array_spec = field_spec["type"]
                elem_spec = array_spec["element"]
                wrap_nested_ptr = self.has_cycle(struct_symbol, elem_spec)
            case "map":
                map_spec = field_spec["type"]
                value_spec = map_spec["value"]
                wrap_nested_ptr = self.has_cycle(struct_symbol, value_spec)
            case _:
                pass
        with self.type_declaration_options(
                wrap_ptr=wrap_ptr,
                wrap_nested_ptr=wrap_nested_ptr
        ):
            field_type = self.get_type_declaration(field_spec["type"])
        with self.nest_name(field_name):
            if field_spec.get("optional", False):
                line = f'{self.indentation()}std::optional<{field_type}> {field_name};'
                if len(line) < self.column_width:
                    self.inline(line, end='\n')
                else:
                    line = f'{self.indentation()}std::optional<{field_type}>'
                    if len(line) < self.column_width:
                        self.inline(line, end='\n')
                        with self.indent():
                            self.write(f'{field_name};')
                    else:
                        self.write('std::optional<')
                        with self.indent():
                            self.write(field_type)
                        self.write(f'> {field_name};')
            else:
                line = f'{self.indentation()}{field_type} {field_name};'
                if len(line) < self.column_width:
                    self.inline(line, end='\n')
                else:
                    self.write(field_type)
                    with self.indent():
                        self.write(f'{field_name};')

    def generate_union_move_assign(self, union_name: str) -> None:
        self.gen_fn_decl(
            'operator=',
            f'{union_name} &',
            params=[
                f'{union_name} &&other'
            ],
        )

    def generate_union_copy_assign(self, union_name: str) -> None:
        self.gen_fn_decl(
            'operator=',
            f'{union_name} &',
            params=[
                f'const {union_name} &other'
            ],
        )

    def generate_union_field_copy_assign(
            self,
            union_name: str,
            field_name: str,
            field_spec: LspSpec
    ) -> None:
        field_type = self.get_type_declaration(field_spec)
        if self.is_union_pointer_type(field_spec):
            param = f'const {field_type} &{field_name}'
        else:
            param = f'{field_type} {field_name}'
        self.gen_fn_decl(
            'operator=',
            f'{union_name} &',
            params=[param],
        )

    def generate_union_field_move_assign(
            self,
            union_name: str,
            field_name: str,
            field_spec: LspSpec
    ) -> None:
        field_type = self.get_type_declaration(field_spec)
        self.gen_fn_decl(
            'operator=',
            f'{union_name} &',
            params=[
                f'{field_type} &&{field_name}'
            ],
        )

    def generate_union_field_getter(
            self,
            field_name: str,
            field_spec: LspSpec) -> None:
        field_type = self.get_type_declaration(field_spec)
        self.gen_fn_decl(field_name, f'const {field_type} &', specs='const')

    def generate_union_tag_getter(self, union_name: str) -> None:
        field_name = 'type'
        enum_name = f'{union_name}Type'
        self.gen_fn_decl(field_name, f'const {enum_name} &', specs='const')

    def generate_union_additional_pointer_setters(
            self,
            union_name: str,
            field_name: str,
            field_spec: LspSpec
    ) -> None:
        with self.type_declaration_options(
                wrap_ptr=True,
                wrap_nested_ptr=False,
                string_ptr=True,
                nested_string_ptr=False
        ):
            self.newline()
            self.generate_union_field_move_assign(union_name, field_name, field_spec)
        with self.type_declaration_options(
                wrap_ptr=False,
                wrap_nested_ptr=False,
                string_ptr=False,
                nested_string_ptr=False
        ):
            self.newline()
            self.generate_union_field_copy_assign(union_name, field_name, field_spec)
            self.newline()
            self.generate_union_field_move_assign(union_name, field_name, field_spec)

    def generate_union(self, union_symbol: LspSymbol) -> None:
        union_spec = union_symbol.spec
        union_name = union_symbol.name
        item_specs = union_spec["items"]
        has_pointer = any((self.is_union_pointer_type(item_spec)
                           for item_spec in item_specs))
        # Generate the field-type enum:
        # -----------------------------
        enum_name = f'{union_name}Type'
        self.gen_enum_with(enum_name, item_specs)
        self.newline()
        # Declare the mapping of field-type enums to their names:
        # -------------------------------------------------------
        names_by_enum = f'{enum_name}Names'
        self.gen_map_decl(names_by_enum, enum_name, 'std::string')
        self.newline()
        union_docs = union_spec.get("documentation", None)
        self.generate_docstring(union_docs)
        # Generate the struct:
        # --------------------
        with self.gen_class(union_name):
            with self.gen_public():
                self.write('// ======================================= //')
                self.write('// Constructor and Destructor Declarations //')
                self.write('// ======================================= //')
                self.newline()
                self.gen_constructor_decl(union_name)
                self.gen_move_constructor_decl(union_name)
                self.gen_copy_constructor_decl(union_name)
                self.gen_destructor_decl(union_name)
                self.newline()
                self.write('// =================== //')
                self.write('// Method Declarations //')
                self.write('// =================== //')
                self.newline()
                self.generate_union_move_assign(union_name)
                self.newline()
                self.generate_union_copy_assign(union_name)
                self.newline()
                self.generate_union_tag_getter(union_name)
                for item_index, item_spec in enumerate(item_specs):
                    with self.nest_name(str(item_index)):
                        field_name = self.name_field(item_spec)
                        with self.type_declaration_options(
                                wrap_ptr=False,
                                wrap_nested_ptr=False,
                                string_ptr=False,
                                nested_string_ptr=False
                        ):
                            self.newline()
                            self.generate_union_field_getter(field_name, item_spec)
                        with self.type_declaration_options(
                                wrap_ptr=True,
                                wrap_nested_ptr=False,
                                string_ptr=True,
                                nested_string_ptr=False
                        ):
                            self.newline()
                            self.generate_union_field_copy_assign(union_name, field_name, item_spec)
                        if self.is_union_pointer_type(item_spec):
                            self.generate_union_additional_pointer_setters(
                                union_name,
                                field_name,
                                item_spec
                            )
            with self.gen_private():
                self.write('// ================== //')
                self.write('// Field Declarations //')
                self.write('// ================== //')
                self.newline()
                self.gen_union_tag_field(enum_name)
                self.newline()
                with self.type_declaration_options(
                        wrap_ptr=True,
                        wrap_nested_ptr=False,
                        string_ptr=True,
                        nested_string_ptr=False,
                ):
                    self.gen_union_with(item_specs)
                if has_pointer:
                    self.newline()
                    self.gen_fn_decl('reset', 'void')
        self.newline()

    def gen_struct_move_assign_decl(self, struct_name: str) -> None:
        self.gen_fn_decl(
            'operator=',
            f'{struct_name} &',
            params=[
                f'{struct_name} &&other'
            ],
        )

    def gen_struct_copy_assign_decl(self, struct_name: str) -> None:
        self.gen_fn_decl(
            'operator=',
            f'{struct_name} &',
            params=[
                f'const {struct_name} &other'
            ],
        )

    def generate_structure(self, struct_symbol: LspSymbol) -> None:
        with self.type_declaration_options(
                wrap_ptr=False,
                wrap_nested_ptr=False,
                string_ptr=False,
                nested_string_ptr=False
        ):
            struct_spec = struct_symbol.spec
            struct_name = struct_symbol.name
            self.generate_docstring(struct_spec.get("documentation", None))
            extends_specs = struct_spec.get("extends", None)
            with self.gen_struct(struct_name, sups=extends_specs):
                self.write('// ======================================= //')
                self.write('// Constructor and Destructor Declarations //')
                self.write('// ======================================= //')
                self.newline()
                self.gen_constructor_decl(struct_name)
                self.gen_move_constructor_decl(struct_name)
                self.gen_copy_constructor_decl(struct_name)
                self.gen_destructor_decl(struct_name)
                self.newline()
                self.write('// =================== //')
                self.write('// Method Declarations //')
                self.write('// =================== //')
                self.newline()
                self.gen_struct_move_assign_decl(struct_name)
                self.newline()
                self.gen_struct_copy_assign_decl(struct_name)
                if struct_symbol.fields is not None and len(struct_symbol.fields) > 0:
                    self.newline()
                    self.write('// ================== //')
                    self.write('// Field Declarations //')
                    self.write('// ================== //')
                    self.newline()
                    for field_spec in struct_symbol.fields:
                        with self.nest_name(field_spec["name"]):
                            self.generate_docstring(
                                field_spec.get("documentation", None)
                            )
                            self.generate_struct_field(struct_symbol, field_spec)
            self.newline()

    def visit_alias_symbol(self, alias_symbol: LspSymbol) -> None:
        with self.nested_name_as(alias_symbol.name):
            with self.type_declaration_options(
                    wrap_ptr=False,
                    wrap_nested_ptr=False,
                    string_ptr=False,
                    nested_string_ptr=False,
            ):
                alias_spec = alias_symbol.spec
                alias_name = alias_spec["name"]
                type_spec = alias_spec["type"]
                type_name = type_spec["name"]
                type_symbol = self.resolve(type_name)
                self.write(f'typedef {type_symbol.resolved_name} {alias_name};')
                self.newline()

    def visit_base_symbol(self, base_symbol: LspSymbol) -> None:
        with self.nested_name_as(base_symbol.name):
            with self.type_declaration_options(
                    wrap_ptr=False,
                    wrap_nested_ptr=False,
                    string_ptr=False,
                    nested_string_ptr=False,
            ):
                base_name = base_symbol.name
                self.write('typedef {cxx_type} {lsp_type};'.format(
                    cxx_type=self.rename_base_type(base_name),
                    lsp_type=rename_type(base_name),
                ))
                self.newline()

    def visit_tuple_symbol(self, tuple_symbol: LspSymbol) -> None:
        with self.nested_name_as(tuple_symbol.name):
            with self.type_declaration_options(
                    wrap_ptr=False,
                    wrap_nested_ptr=False,
                    string_ptr=False,
                    nested_string_ptr=False,
            ):
                tuple_name = tuple_symbol.name
                tuple_spec = tuple_symbol.spec
                tuple_type = self.get_type_declaration(tuple_spec)
                self.write(f'typedef {tuple_type} {tuple_name};')
                self.newline()

    def visit_array_symbol(self, array_symbol: LspSymbol) -> None:
        with self.nested_name_as(array_symbol.name):
            with self.type_declaration_options(
                    wrap_ptr=False,
                    wrap_nested_ptr=False,
                    string_ptr=False,
                    nested_string_ptr=False,
            ):
                array_name = array_symbol.name
                array_spec = array_symbol.spec
                elem_spec = array_spec['element']
                with self.type_declaration_options(
                        wrap_nested_ptr=self.has_cycle(array_symbol, elem_spec)
                ):
                    array_type = self.get_type_declaration(array_spec)
                self.write(f'typedef {array_type} {array_name};')
                self.newline()

    def visit_map_symbol(self, map_symbol: LspSymbol) -> None:
        with self.nested_name_as(map_symbol.name):
            with self.type_declaration_options(
                    wrap_ptr=False,
                    wrap_nested_ptr=False,
                    string_ptr=False,
                    nested_string_ptr=False,
            ):
                map_name = map_symbol.name
                map_spec = map_symbol.spec
                value_spec = map_spec["value"]
                with self.type_declaration_options(
                        wrap_nested_ptr=self.has_cycle(map_symbol, value_spec)
                ):
                    map_type = self.get_type_declaration(map_spec)
                self.write(f'typedef {map_type} {map_name};')
                self.newline()

    def generate_enumeration(self, enum_spec: LspSpec) -> None:
        enum_name = enum_spec["name"]
        lower_enum = lower_first(enum_name)
        upper_enum = upper_first(enum_name)
        type_spec = enum_spec["type"]
        type_name = type_spec["name"]
        normalized_name = self.resolve(type_name).normalized_name
        is_string_type = (normalized_name == "string")
        value_type = self.rename_base_type(type_name)
        self.generate_docstring(enum_spec.get("documentation", None))
        with self.gen_enum(enum_name):
            if is_string_type:
                template = '{enum_name}'
            else:
                template = '{enum_name} = {enum_value}'
            for value in enum_spec["values"]:
                self.generate_docstring(value.get("documentation", None))
                expanded = template.format(
                    enum_name=rename_enum(value["name"]),
                    enum_value=value["value"]
                )
                self.write(f'{expanded},')
        self.newline()
        names_by_enum = f'{enum_name}Names'
        self.gen_map_decl(names_by_enum, enum_name, 'std::string')
        self.newline()
        if is_string_type:
            values_by_enum = f'{enum_name}Values'
            self.gen_map_decl(values_by_enum, enum_name, 'std::string')
            self.newline()
        enum_by_name = f'{lower_enum}ByName'
        self.gen_fn_decl(enum_by_name, enum_name, params=[
            'const std::string &name'
        ])
        self.newline()
        enum_by_value = f'{lower_enum}ByValue'
        if is_string_type:
            param = f'const {value_type} &value'
        else:
            param = f'{value_type} value'
        self.gen_fn_decl(enum_by_value, enum_name, params=[param])
        self.newline()
        if is_string_type:
            is_enum = f'is{upper_enum}'
            param = f'const {value_type} &value'
            self.gen_fn_decl(is_enum, 'bool', params=[param])
            self.newline()


    def recompute_dependency_dependencies(
            self,
            recomputed: Set[LspSymbol],
            symbol: LspSymbol
    ) -> None:
        if symbol.dependencies is not None:
            dependency_extractor = self.pipeline.dependency_extractor
            for dependency in symbol.dependencies:
                if dependency is not symbol and dependency not in recomputed:
                    dependency.dependencies = None
                    dependency_extractor.visit_symbol(dependency)
                    recomputed.add(dependency)
            symbol.dependencies = None
            dependency_extractor.visit_symbol(symbol)
            recomputed.add(symbol)
        else:
            raise ValueError(
                f'Cannot recompute dependency dependencies: symbol.dependencies is None'
            )

    def resolve_cycles(self) -> None:
        dependency_extractor = self.pipeline.dependency_extractor
        blacklist = dependency_extractor.blacklist
        cyclic_dependencies = dependency_extractor.cyclic_dependencies
        for symbol in cyclic_dependencies.keys():
            match symbol.kind:
                case LspSymbolKind.UNION:
                    self.write(f'class {symbol.name};  // Forward-declaration for cyclic dependencies')
                    self.newline()
                    blacklist.add(symbol)
                case LspSymbolKind.STRUCTURE | LspSymbolKind.INNER:
                    self.write(f'struct {symbol.name};  // Forward-declaration for cyclic dependencies')
                    self.newline()
                    blacklist.add(symbol)
        recomputed: Set[LspSymbol] = set()
        cyclic_dependencies.clear()
        for symbol in blacklist:
            self.recompute_dependency_dependencies(recomputed, symbol)
        num_cycles = len(cyclic_dependencies)
        if num_cycles > 0:
            raise RuntimeError(
                'Failed to resolve {length} cycles: [{names}]'.format(
                    length=num_cycles,
                    names=", ".join(map(lambda symbol: symbol.name, cyclic_dependencies))
                )
            )

    def generate_code(self) -> None:
        print(f'Generating: {self.file_path}')

        lsp_version: str = self.schema["metaData"]["version"]

        self.generate_disclaimer()
        self.pragma_once()
        self.newline()
        self.gen_include('cstddef')
        self.gen_include('map')
        self.gen_include('memory')
        self.gen_include('optional')
        self.gen_include('string')
        self.gen_include('tuple')
        self.gen_include('utility')
        self.gen_include('vector')
        self.newline()
        with self.gen_namespace(self.namespace):
            self.write('const std::string JSON_RPC_VERSION = "2.0";')
            self.write(f'const std::string LSP_VERSION = "{lsp_version}";')
            self.newline()
            self.resolve_cycles()
            super().generate_code()
            self.generate_request_enums()
            self.generate_notification_enums()



"""
# Design

This file converts from a GFortran module file representation (documented in
the `gfort_mod_parser.py` module) to an Abstract Semantic Representation (ASR).
"""

# TODO: move this into the lfortran package itself
import sys
sys.path.append("../..")
from lfortran.ast import ast
from lfortran.ast.ast_to_src import ast_to_src
import lfortran.adapters.gfortran.mod as gp

class Type(object):
    pass

class Intrinsic(Type):
    __slots__ = ["kind"]

    def __init__(self, kind=None):
        #super(Intrinsic, self).__init__()
        self.kind = kind

class Integer(Intrinsic):
    def __str__(self):
        return "integer"
class Real(Intrinsic):
    def __str__(self):
        return "real"
class Complex(Intrinsic):
    def __repr__(self):
        return "Complex()"
class String(Intrinsic):
    def __repr__(self):
        return "String()"
class Logical(Intrinsic):
    def __repr__(self):
        return "Logical()"

class Derived(Type):
    def __repr__(self):
        return "Derived()"

class Array(Type):
    __slots__ = ["type", "ndim", "atype", "bounds"]

class Arg:
    __slots__ = ["name", "intent", "type", "symtab", "symidx"]

    def get_decl(self):
        s = "%s, intent(%s) :: %s" % (self.type, self.intent, self.name)
        return s

    def tofortran_arg(self):
        return ast.arg(self.name)

    def tofortran_bound(self, b):
        if isinstance(b, gp.Integer):
            if b.i == 1:
                return ""
            else:
                return str(b.i)
        elif isinstance(b, gp.VarIdx):
            return self.symtab[b.idx].name
        print(b)
        raise NotImplementedError("Unsupported bound type")

# TODO: generate ASR in these functions. The ASR will then get converted
# to AST and to Fortran code. This will simplify all the declaration code
# below, which is implicit in the ASR's symbol table.

    def tofortran_decl(self):
        attrs = [
            ast.Attribute(name="intent",
                    args=[ast.attribute_arg(arg=self.intent)]),
        ]
        if isinstance(self.type, Array):
            stype = str(self.type.type)
            args = []
            for i in range(self.type.ndim):
                if self.type.atype == "explicit_shape":
                    s = self.tofortran_bound(self.type.bounds[i][0])
                    if s != "":
                        s += ":"
                    s += self.tofortran_bound(self.type.bounds[i][1])
                    args.append(s)
                elif self.type.atype == "assumed_shape":
                    s = self.tofortran_bound(self.type.bounds[i][0])
                    s += ":"
                    args.append(s)
                else:
                    assert False
            args = [ast.attribute_arg(arg=x) for x in args]
            attrs.append(ast.Attribute(name="dimension", args=args))
        else:
            stype = str(self.type)
        decl = ast.decl(sym=self.name, sym_type=stype,
            dims=[], attrs=attrs)
        return ast.Declaration(vars=[decl], lineno=1, col_offset=1)

    def tofortran_cdecl(self):
        attrs = [
            ast.Attribute(name="intent",
                    args=[ast.attribute_arg(arg=self.intent)]),
        ]
        if isinstance(self.type, Array):
            args = []
            if self.type.atype == "assumed_shape":
                if self.type.ndim == 1:
                    stype = "type(c_desc1_t)"
                elif self.type.ndim == 2:
                    stype = "type(c_desc2_t)"
                else:
                    raise NotImplementedError("Assumed shape dim")
            else:
                stype = str(self.type.type)
                for i in range(self.type.ndim):
                    if self.type.atype == "explicit_shape":
                        s = self.tofortran_bound(self.type.bounds[i][0])
                        if s != "":
                            s += ":"
                        s += self.tofortran_bound(self.type.bounds[i][1])
                        args.append(s)
                    else:
                        assert False
                args = [ast.attribute_arg(arg=x) for x in args]
                attrs.append(ast.Attribute(name="dimension", args=args))
        else:
            stype = str(self.type)
        decl = ast.decl(sym=self.name, sym_type=stype,
            dims=[], attrs=attrs)
        return ast.Declaration(vars=[decl], lineno=1, col_offset=1)

class Function:
    __slots__ = ["name", "args", "return_type", "mangled_name"]

    def tofortran_impl(self):
        args = [x.tofortran_arg() for x in self.args]
        args_decl = [x.tofortran_decl() for x in self.args]
        iface_decl = [
            ast.Interface2(name="x", procs=[self.tofortran_iface()], lineno=1, col_offset=1)
        ]
        return_var = ast.Name(id="r", lineno=1, col_offset=1)
        cname = self.name + "_c_wrapper"
        cargs = []
        for x in self.args:
            a = ast.Name(id=x.name, lineno=1, col_offset=1)
            if isinstance(x.type, Array) and x.type.atype == "assumed_shape":
                cargs.append(
                    ast.FuncCallOrArray(
                        func="c_desc", args=[a], keywords=[],
                        lineno=1, col_offset=1
                    )
                )
            else:
                cargs.append(a)
        body = [
            ast.Assignment(
                target=return_var,
                value=ast.FuncCallOrArray(
                    func=cname, args=cargs, keywords=[],
                    lineno=1, col_offset=1
                ),
                lineno=1, col_offset=1
            )
        ]
        return ast.Function(name=self.name, args=args,
            return_type=str(self.return_type), return_var=return_var, bind=None,
            use=[],
            decl=args_decl+iface_decl, body=body, contains=[], lineno=1,
            col_offset=1)

    def tofortran_iface(self):
        args = [x.tofortran_arg() for x in self.args]
        args_decl = [x.tofortran_cdecl() for x in self.args]
        use = [
            ast.Use(module="gfort_interop",
                symbols=[
                    ast.UseSymbol(sym="c_desc1_t", rename=None),
                    ast.UseSymbol(sym="c_desc2_t", rename=None),
                ], lineno=1, col_offset=1)
        ]
        cname = self.name + "_c_wrapper"
        bind_args = [ast.keyword(arg=None, value=ast.Name(id="c", lineno=1, col_offset=1)),
            ast.keyword(arg="name", value=ast.Str(s=self.mangled_name, lineno=1, col_offset=1))]
        bind = ast.Bind(args=bind_args)
        return ast.Function(name=cname, args=args,
            return_type=str(self.return_type), return_var=None,
            bind=bind, use=use,
            decl=args_decl, body=[], contains=[], lineno=1, col_offset=1)

class Module:
    __slots__ = ["name", "contains"]

    def tofortran(self):
        contains = [x.tofortran_impl() for x in self.contains]
        use = [
            ast.Use(module="gfort_interop",
                symbols=[
                    ast.UseSymbol(sym="c_desc", rename=None),
                ], lineno=1, col_offset=1)
        ]
        return ast.Module(name=self.name, use=use, decl=[], contains=contains)


def convert_arg(table, idx):
    arg = table[idx]
    a = Arg()
    assert isinstance(arg.name, str)
    a.name = arg.name
    assert isinstance(arg.intent, str)
    a.intent = arg.intent
    assert isinstance(table, dict)
    a.symtab = table
    assert isinstance(idx, int)
    a.symidx = idx
    assert isinstance(arg.type, str)
    type_ = arg.type
    if arg.bounds:
        ar = Array()
        ar.type = type_
        ar.ndim = len(arg.bounds)
        if arg.bounds[0][1] is None:
            ar.atype = "assumed_shape"
        else:
            ar.atype = "explicit_shape"
        ar.bounds = arg.bounds
        a.type = ar
    else:
        a.type = type_
    return a

def convert_function(table, f):
    assert isinstance(f, gp.Procedure)
    fn = Function()
    assert isinstance(f.name, str)
    fn.name = f.name
    assert isinstance(f.type, str)
    fn.return_type = f.type
    fn.mangled_name = '__' + f.mod + '_MOD_' + f.name.lower()
    args = []
    for arg in f.args:
        assert isinstance(arg, gp.VarIdx)
        args.append(convert_arg(table, arg.idx))
    fn.args = args
    return fn

def convert_module(table, public_symbols):
    m = Module()
    contains = []
    module_name = None
    for sym in public_symbols:
        s = table[sym.idx.idx]
        if isinstance(s, gp.Module):
            # Skip modules if they are listed in public symbols
            continue
        assert isinstance(s, gp.Procedure)
        if module_name:
            assert module_name == s.mod
        else:
            module_name = s.mod
        contains.append(convert_function(table, s))
    m.name = module_name
    m.contains = contains
    return m

version, orig_file, table, public_symbols = gp.load_module("mod1.mod")
m = convert_module(table, public_symbols)
a = m.tofortran()
a.name = "mod2"
s = ast_to_src(a)
print(s)



import re
from collections import defaultdict, deque
from contextlib import contextmanager
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Union

from llanguage_server.cxx.lsp_file_generator import gensym_context
from llanguage_server.cxx.visitors import BaseCPlusPlusLspVisitor
from llanguage_server.lsp.datatypes import LspSpec, LspSymbol
from llanguage_server.lsp.utils import (any_enum, lower_first,
                                        method_to_camel_case, rename_enum,
                                        rename_field, rename_type, upper_first)
from llanguage_server.lsp.visitors import (LspAnalysisPipeline, LspSymbol,
                                           LspSymbolKind)

RE_VOWEL: re.Pattern = re.compile(r'^[aeiouy]', re.IGNORECASE)


SpecTypesAndTypeSpecs = List[
    Tuple[str, LspSpec]
]

MapTypesAndTypeSpecs = List[
    Tuple[
        str,
        LspSpec,
        Dict[str, List[Any]],
        Dict[str, List[Any]]
    ]
]

ArraySpecTypesAndTypeSpecs = List[
    Tuple[
        str,
        LspSpec,
        Dict[str, List[Any]]
    ]
]


def type_symbol_or_spec_or_name_fn(fn: Callable) -> Callable:
    memo = {}
    @wraps(fn)
    def wrapper(
            self,
            type_symbol_or_spec_or_name: Union[LspSymbol, LspSpec, str],
            *args,
            **kwargs
    ):
        match type_symbol_or_spec_or_name:
            case LspSymbol():
                type_symbol: LspSymbol = type_symbol_or_spec_or_name
                type_name = type_symbol.name
            case dict():
                type_spec: LspSpec = type_symbol_or_spec_or_name
                type_name = self.get_type_name(type_spec)
            case str():
                type_name: str = type_symbol_or_spec_or_name
        if type_name in memo:
            return memo[type_name]
        retval = fn(self, type_name, *args, **kwargs)
        memo[type_name] = retval
        return retval
    return wrapper


class CPlusPlusLspTransformerSourceGenerator(BaseCPlusPlusLspVisitor):

    def __init__(
            self,
            output_dir: Path,
            schema: LspSpec,
            pipeline: LspAnalysisPipeline,
            namespace: str
    ) -> None:
        specification_source = output_dir / "lsp_transformer.cpp"
        super().__init__(specification_source, schema, pipeline, namespace)

    def gen_as_any(self, value: str, move: bool = True) -> str:
        any_name = self.gensym_init('any', 'std::make_unique<LSPAny>()')
        expr = f'std::move({value})' if move else value
        self.gen_assign(f'(*{any_name})', expr)
        return any_name

    def move_as_any(self, expr: str, move_expr: bool = True) -> str:
        any_name = self.gen_as_any(expr, move=move_expr)
        return f'std::move({any_name})'

    def gen_return_as_any(self, value: str, move: bool = True) -> None:
        any_name = self.gensym_decl('LSPAny', 'any')
        expr = f'std::move({value})' if move else value
        self.gen_assign(any_name, expr)
        self.gen_return(any_name)

    def article_for(self, name: str) -> str:
        return 'an' if RE_VOWEL.match(name) else 'a'

    @contextmanager
    def gen_any_to_type(
            self,
            type_symbol_or_spec_or_name: Union[LspSymbol, LspSpec, str],
            return_type_or_spec: Optional[Union[str, LspSpec]] = None
    ) -> Iterator:
        return_type: str
        match return_type_or_spec:
            case None:
                match type_symbol_or_spec_or_name:
                    case LspSymbol():
                        type_symbol: LspSymbol = type_symbol_or_spec_or_name
                        return_type = type_symbol.name
                    case dict():
                        type_spec: LspSpec = type_symbol_or_spec_or_name
                        return_type = self.get_type_name(type_spec)
                    case str():
                        return_type = type_symbol_or_spec_or_name
                    case _:
                        raise ValueError(
                            'Unsupported return type ({kind}): {spec}'.format(
                                kind=type(type_symbol_or_spec_or_name),
                                spec=type_symbol_or_spec_or_name,
                            )
                        )
            case dict():
                return_spec: LspSpec = return_type_or_spec
                return_type = self.get_type_declaration(return_spec)
            case str():
                return_type = return_type_or_spec
            case _:
                raise ValueError(
                    'Unsupported return type ({kind}): {spec}'.format(
                        kind=type(return_type_or_spec),
                        spec=return_type_or_spec,
                    )
                )
        any_to_type = self.get_any_to_type_name(type_symbol_or_spec_or_name)
        with self.gen_fn(
            f'LspTransformer::{any_to_type}',
            return_type,
            params=[
                'const LSPAny &any',
            ],
            specs='const'
        ):
            yield self
        self.newline()

    @contextmanager
    def gen_type_to_any(
            self,
            type_symbol_or_spec_or_name: Union[LspSymbol, LspSpec, str],
            type_param: str
    ) -> Iterator:
        type_to_any = self.get_type_to_any_name(type_symbol_or_spec_or_name)
        with self.gen_fn(
            f'LspTransformer::{type_to_any}',
            'LSPAny',
            params=[
                type_param,
            ],
            specs='const'
        ):
            yield self
        self.newline()

    def value_of(self, value: str, is_optional: bool) -> str:
        if is_optional:
            return f'{value}.value()'
        return value

    def deref(
            self,
            symbol_or_spec_or_name: Union[LspSymbol, LspSpec, str],
            value: str
    ) -> str:
        name: str
        symbol: LspSymbol
        match symbol_or_spec_or_name:
            case LspSymbol():
                symbol = symbol_or_spec_or_name
            case dict():
                spec: LspSpec = symbol_or_spec_or_name
                if "name" in spec:
                    name = spec["name"]
                else:
                    name = self.nested_name()
                symbol = self.resolve(name)
            case str():
                name = symbol_or_spec_or_name
                symbol = self.resolve(name)
            case _:
                raise ValueError(
                    'Unsupported specification type ({kind}): {spec}'.format(
                        kind=type(symbol_or_spec_or_name),
                        spec=symbol_or_spec_or_name,
                    )
                )
        if symbol.resolution is not None:
            symbol = symbol.resolution
        match symbol.kind:
            case LspSymbolKind.BASE:
                if symbol.normalized_name == "string":
                    return f'*{value}'
                return value
            case _:
                return f'*{value}'

    @type_symbol_or_spec_or_name_fn
    def get_any_to_type_name(self, type_name: str) -> str:
        upper_type = upper_first(type_name)
        return f'anyTo{upper_type}'

    @type_symbol_or_spec_or_name_fn
    def get_type_to_any_name(self, type_name: str) -> str:
        lower_type = lower_first(type_name)
        return f'{lower_type}ToAny'

    @gensym_context
    def gen_any_to_enum(self, enum_spec: LspSpec) -> None:
        enum_name = enum_spec["name"]
        type_spec = enum_spec["type"]
        type_name = type_spec["name"]
        value_type = rename_type(type_name)
        field_name = self.name_field(type_spec)
        lower_enum = lower_first(enum_name)
        enum_by_value = f'{lower_enum}ByValue'
        enumerator = rename_enum(type_name)
        with self.gen_any_to_type(enum_name):
            with self.gen_try():
                with self.gen_switch('any.type()'):
                    match type_name:
                        case "string":
                            with self.gen_case('LSPAnyType', 'string'):
                                self.gen_assign(f'const {value_type} &value', f'any.{field_name}()')
                                self.gen_return(f'{enum_by_value}(value)')
                        case "integer":
                            with self.gen_case('LSPAnyType', 'integer'):
                                self.gen_assign(f'{value_type} value', f'any.{field_name}()')
                                self.gen_return(f'{enum_by_value}(value)')
                            with self.gen_case('LSPAnyType', 'uinteger'):
                                self.gen_assign(
                                    f'{value_type} value',
                                    f'static_cast<{value_type}>(any.{field_name}())'
                                )
                                self.gen_return(f'{enum_by_value}(value)')
                        case "uinteger":
                            with self.gen_case('LSPAnyType', 'uinteger'):
                                self.gen_assign(f'{value_type} value', f'any.{field_name}()')
                                self.gen_return(f'{enum_by_value}(value)')
                            with self.gen_case('LSPAnyType', 'integer'):
                                self.gen_assign(
                                    f'{value_type} value',
                                    f'static_cast<{value_type}>(any.{field_name}())'
                                )
                                self.gen_return(f'{enum_by_value}(value)')
                        case _:
                            raise ValueError(f'Unsupported enumeration type ({type_name}): {enum_spec}')
                    with self.gen_default():
                        self.gen_throw_invalid_params(
                            f'("LSPAnyType for {self.article_for(enum_name)} "',
                            f' "{enum_name}"',
                            f' " must be of type LSPAnyType::{enumerator}"',
                            f' " but received LSPAnyType::" + LSPAnyTypeNames.at(any.type()))'
                        )
            with self.gen_catch('std::invalid_argument &e'):
                self.gen_throw_invalid_params('e.what()')

    @gensym_context
    def gen_enum_to_any(self, enum_spec: LspSpec) -> None:
        enum_name = enum_spec["name"]
        type_spec = enum_spec["type"]
        type_name = type_spec["name"]
        value_type = rename_type(type_name)
        enum_param =f'{enum_name} enumerator'
        with self.gen_type_to_any(enum_name, enum_param):
            any_name = self.gensym_decl('LSPAny', 'any')
            if self.resolve(type_name).normalized_name == "string":
                values_by_enum = f'{enum_name}Values'
                self.gen_assign(any_name, f'{values_by_enum}.at(enumerator)')
            else:
                self.gen_assign(any_name, f'static_cast<{value_type}>(enumerator)')
            self.gen_return(any_name)

    def generate_enumeration(self, enum_spec: LspSpec) -> None:
        self.gen_any_to_enum(enum_spec)
        self.gen_enum_to_any(enum_spec)

    def get_type_name(self, type_spec: LspSpec) -> str:
        match type_spec["kind"]:
            case "base" | "reference" | "enumeration":
                return type_spec["name"]
            case "literal" | "or" | "array" | "map" | "tuple":
                return self.nested_name()
            case _:
                raise ValueError(f'Cannot determine type name for: {type_spec}')

    def get_upper_type_name(self, type_spec: LspSpec) -> str:
        type_name = self.get_type_name(type_spec)
        return upper_first(type_name)

    def generate_upper_type_name(self, type_spec: LspSpec) -> None:
        self.inline(self.get_upper_type_name(type_spec))

    def get_lower_type_name(self, type_spec: LspSpec) -> str:
        type_name = self.get_type_name(type_spec)
        return lower_first(type_name)

    def generate_lower_type_name(self, type_spec: LspSpec) -> None:
        self.inline(self.get_lower_type_name(type_spec))

    def index_by_type(
            self,
            type_index: Dict[str, List[Any]],
            type_spec: LspSpec,
            level: int = 0,
            type_name: Optional[str] = None
    ) -> None:
        type_kind = type_spec["kind"]
        match type_kind:
            case "base":
                base_spec = type_spec
                base_name = base_spec["name"]
                if self.resolve(base_name).normalized_name == "string":
                    index = type_index["string"]
                else:
                    index = type_index[base_name]
                index.append(("base", base_spec))
            case "reference":
                if type_name is None:
                    type_name = type_spec["name"]
                symbol = self.resolve(type_name)
                with self.nested_name_as(symbol.name):
                    match type_name:
                        case "LSPAny" | "LSPObject":
                            index = type_index["structure"]
                            index.append((type_name, symbol.spec))
                        case _:
                            match symbol.kind:
                                case LspSymbolKind.STRUCTURE:
                                    struct_spec = type_spec
                                    if type_name is not None:
                                        struct_name = type_name
                                    else:
                                        struct_name = type_spec["name"]
                                    index = type_index["structure"]
                                    index.append((struct_name, struct_spec))
                                case LspSymbolKind.ENUMERATION:
                                    enum_spec = type_spec
                                    if type_name is not None:
                                        enum_name = type_name
                                    else:
                                        enum_name = enum_spec["name"]
                                    index = type_index["enumeration"]
                                    index.append((enum_name, enum_spec))
                                case _:
                                    self.index_by_type(type_index, symbol.spec, level, type_name)
            case "array":
                array_spec = type_spec
                elem_spec = array_spec["element"]
                elem_index = defaultdict(list)
                with self.nest_name('elem'):
                    self.index_by_type(elem_index, elem_spec, 1 + level)
                index = type_index["array"]
                array_name = self.nested_name()
                index.append((array_name, type_spec, elem_index))
            case "map":
                key_spec = type_spec["key"]
                key_index = defaultdict(list)
                with self.nest_name('key'):
                    self.index_by_type(key_index, key_spec, 1 + level)
                value_spec = type_spec["value"]
                value_index = defaultdict(list)
                with self.nest_name('value'):
                    self.index_by_type(value_index, value_spec, 1 + level)
                index = type_index["map"]
                map_name = self.nested_name()
                index.append((map_name, type_spec, key_index, value_index))
            case "and":
                raise ValueError(f'AND types are not supported: {type_spec}')
            case "or":
                union_spec = type_spec
                if level == 0:
                    for i, item_spec in enumerate(union_spec["items"]):
                        with self.nest_name(str(i)):
                            self.index_by_type(type_index, item_spec, 1 + level)
                else:
                    if type_name is None:
                        type_name = self.nested_name()
                    index = type_index['union']
                    index.append((type_name, union_spec))
            case "tuple":
                item_indices = []
                for i, item_spec in enumerate(type_spec["items"]):
                    with self.nest_name(str(i)):
                        item_index = defaultdict(list)
                        self.index_by_type(item_index, item_spec, 1 + level)
                        item_indices.append(item_index)
                index = type_index[type_kind]
                index.append((self.nested_name(), type_spec, item_indices))
            case "literal":
                index = type_index["inner"]
                if type_name is not None:
                    inner_name = type_name
                else:
                    inner_name = self.nested_name()
                index.append((inner_name, type_spec))
            case "stringLiteral":
                index = type_index["string"]
                index.append(type_spec)
            case "integerLiteral":
                index = type_index["integer"]
                index.append(type_spec)
            case "booleanLiteral":
                index = type_index["boolean"]
                index.append(type_spec)
            case _:
                raise ValueError(
                    f'Unsupported index type ({type_kind}): {type_spec}'
                )

    def gen_rec_any_to_union_type(
            self,
            union_name: str,
            value_name: str,
            type_name: str,
            spec_types_and_type_specs: SpecTypesAndTypeSpecs
    ) -> None:
        if len(spec_types_and_type_specs) > 0:
            spec_type, type_spec = spec_types_and_type_specs[0]
            if spec_type == "enumeration":
                with self.gen_try():
                    enum_spec = type_spec
                    enum_name = enum_spec["name"]
                    upper_enum = upper_first(enum_name)
                    any_to_enum = f'anyTo{upper_enum}'
                    self.gen_assign(f'{value_name}', f'{any_to_enum}(any)')
                with self.gen_catch('LspException &/*e*/'):
                    self.gen_rec_any_to_union_type(
                        union_name,
                        value_name,
                        type_name,
                        spec_types_and_type_specs[1:]
                    )
            elif len(spec_types_and_type_specs) == 1:
                upper_name = upper_first(type_name)
                any_to_type = f'anyTo{upper_name}'
                self.gen_assign(f'{value_name}', f'{any_to_type}(any)')
            else:
                raise ValueError(f'Redundant {type_name} specs detected')
        else:
            self.gen_throw_invalid_params(
                f'"Failed to transform LSPAny to {union_name}"'
            )

    def gen_rec_any_to_union_inner(
            self,
            union_name: str,
            value_name: str,
            struct_specs: Optional[SpecTypesAndTypeSpecs],
            inner_specs: Optional[SpecTypesAndTypeSpecs],
            union_specs: Optional[SpecTypesAndTypeSpecs],
            map_specs: Optional[MapTypesAndTypeSpecs],
            array_specs: Optional[ArraySpecTypesAndTypeSpecs]
    ) -> None:
        type_name: Optional[str] = None

        if struct_specs is not None and len(struct_specs) > 0:
            type_name, _ = struct_specs[0]
            struct_specs = struct_specs[1:]
        elif inner_specs is not None and len(inner_specs) > 0:
            type_name, _ = inner_specs[0]
            inner_specs = inner_specs[1:]
        elif union_specs is not None and len(union_specs) > 0:
            type_name, _ = union_specs[0]
            union_specs = union_specs[1:]
        elif map_specs is not None and len(map_specs) > 0:
            type_name, _, _, _ = map_specs[0]
            map_specs = map_specs[1:]
        elif array_specs is not None and len(array_specs) > 0:
            type_name, _, _ = array_specs[0]
            array_specs = array_specs[1:]

        if type_name is not None:
            with self.gen_try():
                any_to_type = self.get_any_to_type_name(type_name)
                self.gen_assign(f'{value_name}', f'{any_to_type}(any)')
            with self.gen_catch('LspException &/*e*/'):
                self.gen_rec_any_to_union_inner(
                    union_name,
                    value_name,
                    struct_specs,
                    inner_specs,
                    union_specs,
                    map_specs,
                    array_specs
                )
        else:
            self.gen_throw_invalid_params(
                f'"Failed to transform LSPAny to {union_name}"'
            )

    @gensym_context
    def generate_any_to_union(self, union_symbol: LspSymbol) -> None:
        union_spec = union_symbol.spec
        union_name = union_symbol.name
        type_index = defaultdict(list)
        self.index_by_type(type_index, union_spec)
        struct_specs = type_index.get("structure", None)
        inner_specs = type_index.get("inner", None)
        union_specs = type_index.get("union", None)
        map_specs = type_index.get("map", None)
        has_object_type = (struct_specs is not None) \
            or (inner_specs is not None) \
            or (union_specs is not None) \
            or (map_specs is not None)
        with self.gen_any_to_type(union_name, union_name):
            value_name = self.gensym_decl(union_name, 'value')
            self.newline()
            with self.gen_switch('any.type()'):
                if has_object_type:
                    with self.gen_case('LSPAnyType', 'object'):
                        self.gen_rec_any_to_union_inner(
                            union_name, value_name,
                            struct_specs, inner_specs, union_specs, map_specs,
                            None
                        )
                        self.gen_break()
                array_specs = type_index.get("array", None)
                if array_specs is not None:
                    with self.gen_case('LSPAnyType', 'array'):
                        self.gen_rec_any_to_union_inner(
                            union_name, value_name,
                            None, None, None, None,
                            array_specs
                        )
                        self.gen_break()
                string_specs = type_index.get("string", None)
                if string_specs is not None:
                    string_specs.sort(key=lambda pair: pair[0])
                    with self.gen_case('LSPAnyType', 'string'):
                        self.gen_rec_any_to_union_type(
                            union_name,
                            value_name,
                            'string',
                            string_specs
                        )
                        self.gen_break()
                integer_specs = type_index.get("integer", None)
                if integer_specs is not None:
                    with self.gen_case('LSPAnyType', 'integer'):
                        self.gen_rec_any_to_union_type(
                            union_name,
                            value_name,
                            'integer',
                            integer_specs
                        )
                        self.gen_break()
                uinteger_specs = type_index.get("uinteger", None)
                if uinteger_specs is not None:
                    with self.gen_case('LSPAnyType', 'uinteger'):
                        self.gen_rec_any_to_union_type(
                            union_name,
                            value_name,
                            'uinteger',
                            uinteger_specs
                        )
                        self.gen_break()
                decimal_specs = type_index.get("decimal", None)
                if decimal_specs is not None:
                    with self.gen_case('LSPAnyType', 'decimal'):
                        self.gen_assign(f'{value_name}', f'anyToDecimal(any)')
                        self.gen_break()
                boolean_specs = type_index.get("boolean", None)
                if boolean_specs is not None:
                    with self.gen_case('LSPAnyType', 'boolean'):
                        self.gen_assign(f'{value_name}', f'anyToBoolean(any)')
                        self.gen_break()
                null_specs = type_index.get("null", None)
                if null_specs is not None:
                    with self.gen_case('LSPAnyType', 'null'):
                        self.gen_assign(f'{value_name}', f'anyToNull(any)')
                        self.gen_break()
                with self.gen_default():
                    self.gen_throw_invalid_params(
                        f'("Invalid LSPAnyType for {self.article_for(union_name)} "',
                        f' "{union_name}"',
                        f' ": " + LSPAnyTypeNames.at(any.type()))'
                    )
            self.newline()
            self.gen_return(f'{value_name}')

    @gensym_context
    def generate_union_to_any(self, union_symbol: LspSymbol) -> None:
        union_spec = union_symbol.spec
        union_name = union_symbol.name
        enum_name = f'{union_name}Type'
        with self.gen_type_to_any(union_name, f'const {union_name} &value'):
            with self.gen_switch('value.type()'):
                for i, item_spec in enumerate(union_spec["items"]):
                    with self.nest_name(str(i)):
                        field_name = self.name_field(item_spec)
                        with self.gen_case(enum_name, item_spec):
                            match item_spec["kind"]:
                                case "array":
                                    array_spec = item_spec
                                    elem_spec = array_spec["element"]
                                    array_name = self.gensym_array(
                                        "LSPArray", "array",
                                        f'value.{field_name}().size()'
                                    )
                                    with self.gensym_foreach(f'value.{field_name}()') as elem_name:
                                        with self.nest_name('elem'):
                                            elem_to_any = self.get_type_to_any_name(elem_spec)
                                        elem_expr = f'{elem_to_any}({elem_name})'
                                        self.gen_call(
                                            f'{array_name}.push_back',
                                            self.move_as_any(elem_expr, move_expr=False)
                                        )
                                    any_name = self.gensym_decl('LSPAny', 'any')
                                    self.gen_assign(any_name, f'std::move({array_name})')
                                    self.gen_return(any_name)
                                case "tuple":
                                    tuple_spec = item_spec
                                    item_specs = tuple_spec["items"]
                                    array_name = self.gensym_array(
                                        "LSPArray", "array",
                                        str(len(item_specs))
                                    )
                                    tuple_type = self.get_type_declaration(tuple_spec)
                                    tuple_name = self.gensym('tuple')
                                    self.gen_assign(
                                        f'const {tuple_type} &{tuple_name}',
                                        f'value.{field_name}()'
                                    )
                                    for i, item_spec in enumerate(item_specs):
                                        lower_item = self.get_lower_type_name(item_spec)
                                        item_to_any = f'{lower_item}ToAny'
                                        elem_expr = f'{item_to_any}(std::get<{i}>({tuple_name}))'
                                        self.gen_call(
                                            f'{array_name}.push_back',
                                            self.move_as_any(elem_expr, move_expr=False)
                                        )
                                    any_name = self.gensym_decl('LSPAny', 'any')
                                    self.gen_assign(any_name, f'std::move({array_name})')
                                    self.gen_return(any_name)
                                case "reference":
                                    item_type_name = item_spec["name"]
                                    match item_type_name:
                                        case "LSPAny":
                                            self.gen_return(f'copy(value.{field_name}())')
                                        case "LSPObject" | "LSPArray":
                                            any_name = self.gensym_decl('LSPAny', 'any')
                                            self.gen_assign(any_name, f'copy(value.{field_name}())')
                                            self.gen_return(any_name)
                                        case _:
                                            lower_item = self.get_lower_type_name(item_spec)
                                            item_to_any = f'{lower_item}ToAny'
                                            self.gen_return(f'{item_to_any}(value.{field_name}())')
                                case _:
                                    lower_item = self.get_lower_type_name(item_spec)
                                    item_to_any = f'{lower_item}ToAny'
                                    self.gen_return(f'{item_to_any}(value.{field_name}())')
                with self.gen_default():
                    self.gen_throw_invalid_params(
                        f'("Unsupported {union_name}Type: " +',
                        f' {union_name}TypeNames.at(value.type()))'
                    )

    def generate_union(self, union_symbol: LspSymbol) -> None:
        self.generate_any_to_union(union_symbol)
        self.generate_union_to_any(union_symbol)

    def expand_fields(self, struct_spec_or_symbol_or_name: Union[LspSpec, LspSymbol, str]) -> Iterator[Tuple[str, LspSpec]]:
        struct_name: str
        struct_spec: LspSpec
        struct_symbol: LspSymbol
        match struct_spec_or_symbol_or_name:
            case dict():
                struct_spec = struct_spec_or_symbol_or_name
                if "name" in struct_spec:
                    struct_name = struct_spec["name"]
                else:
                    struct_name = self.nested_name()
                struct_symbol = self.resolve(struct_name)
                struct_spec = struct_symbol.spec
            case LspSymbol():
                struct_symbol = struct_spec_or_symbol_or_name
                struct_spec = struct_symbol.spec
                struct_name = struct_symbol.name
            case str():
                struct_name = struct_spec_or_symbol_or_name
                struct_symbol = self.resolve(struct_name)
                struct_spec = struct_symbol.spec
        extends_specs: Optional[List[LspSpec]] = struct_spec.get("extends", None)
        if extends_specs is not None:
            for extends_spec in extends_specs:
                super_name = extends_spec['name']
                super_symbol = self.resolve(super_name)
                yield from self.expand_fields(super_symbol)
        mixin_specs: Optional[List[LspSpec]] = struct_spec.get("mixins", None)
        if mixin_specs is not None:
            for mixin_spec in mixin_specs:
                mixin_name = mixin_spec['name']
                mixin_symbol = self.resolve(mixin_name)
                if mixin_symbol.resolution is not None:
                    mixin_symbol = mixin_symbol.resolution
                yield from self.expand_fields(mixin_symbol)
        if struct_symbol.fields is not None:
            for field_spec in struct_symbol.fields:
                yield struct_name, field_spec

    @gensym_context
    def generate_any_to_struct(self, struct_symbol: LspSymbol) -> None:
        struct_spec = struct_symbol.spec
        if "name" in struct_spec:
            struct_name = struct_spec["name"]
        else:
            struct_name = self.nested_name()
        with self.gen_any_to_type(struct_name, struct_name):
            enum_type = f'LSPAnyType::{rename_enum("object")}'
            with self.gen_if_ne('any.type()', enum_type):
                self.gen_throw_invalid_params(
                    f'("LSPAnyType for {self.article_for(struct_name)} "',
                    f' "{struct_name}"',
                    f' " must be of type {enum_type}"',
                    f' " but received LSPAnyType::" + LSPAnyTypeNames.at(any.type()))'
                )
            self.newline()
            value_name = self.gensym_decl(struct_name, 'value')
            self.newline()
            array_field = rename_field("array")
            object_field = rename_field("object")
            object_name = self.gensym_ref('object', f'any.{object_field}()')
            iter_name = self.gensym_decl('LSPObject::const_iterator', 'iter')
            type_names_and_field_specs = list(self.expand_fields(struct_spec))
            self.newline()
            num_fields = len(type_names_and_field_specs)
            with self.gen_if(f'{object_name}.size() > {num_fields}'):
                self.gen_throw_invalid_params(
                    f'("Too many attributes to transform to {self.article_for(struct_name)} "',
                    f' "{struct_name}"',
                    f' ": " + std::to_string({object_name}.size()) + " > {num_fields}")'
                )
            for field_type_name, field_spec in type_names_and_field_specs:
                field_name = field_spec["name"]
                with self.nested_names_as(deque([field_type_name, field_name])):
                    field_type_spec = field_spec["type"]
                    self.newline()
                    self.gen_assign(f'{iter_name}', f'{object_name}.find("{field_name}")')
                    with self.gen_if_ne(iter_name, f'{object_name}.end()', end=''):
                        match field_type_spec["kind"]:
                            case "base":
                                any_to_base = self.get_any_to_type_name(field_spec["type"])
                                field_expr = f'{any_to_base}(*{iter_name}->second)'
                            case "reference":
                                ref_type_name = self.get_type_name(field_spec["type"])
                                match ref_type_name:
                                    case "LSPAny":
                                        field_expr = f'copy(*{iter_name}->second)'
                                    case "LSPObject":
                                        field_expr = f'copy({iter_name}->second->{object_field}())'
                                    case _:
                                        any_to_ref = f'anyTo{ref_type_name}'
                                        field_expr = f'{any_to_ref}(*{iter_name}->second)'
                            case "literal" | "or":
                                nested_name = self.nested_name([field_type_name, field_name])
                                # nested_name = self.nested_name()
                                any_to_nested = self.get_any_to_type_name(nested_name)
                                field_expr = f'{any_to_nested}(*{iter_name}->second)'
                            case "array":
                                array_spec = field_type_spec
                                elem_spec = array_spec["element"]
                                array_name = self.gensym_ref('array', f'{iter_name}->second->{array_field}()')
                                values_name = self.gensym_decl(array_spec, 'values')
                                with self.nest_name('elem'):
                                    any_to_elem = self.get_any_to_type_name(elem_spec)
                                with self.gensym_foreach(array_name) as elem_name:
                                    match elem_spec["kind"]:
                                        case "base":
                                            array_expr = f'{any_to_elem}(*{elem_name})'
                                        case "reference":
                                            match elem_spec["name"]:
                                                case "LSPAny":
                                                    array_expr = f'copy(*{elem_name})'
                                                case "LSPObject":
                                                    array_expr = f'{elem_name}->{object_field}()'
                                                case _:
                                                    array_expr = f'{any_to_elem}(*{elem_name})'
                                        case "or" | "literal":
                                            # nested_name = self.nested_name([field_type_name, field_name])
                                            with self.nest_name('elem') as nested_name:
                                                if nested_name == "LSPAny":
                                                    array_expr = f'copy(*{elem_name})'
                                                else:
                                                    any_to_nested = self.get_any_to_type_name(nested_name)
                                                    array_expr = f'{any_to_nested}(*{elem_name})'
                                        case _:
                                            raise ValueError(
                                                f'Unsupported array type ({elem_spec["kind"]}): {elem_spec}'
                                            )
                                    if self.has_cycle(struct_symbol, elem_spec):
                                        elem_type = self.get_type_declaration(elem_spec)
                                        array_expr = f'std::make_unique<{elem_type}>({array_expr})'
                                    self.gen_call(f'{values_name}.push_back', array_expr)
                                field_expr = f'std::move({values_name})'
                            case "map":
                                map_spec = field_type_spec
                                value_spec = field_type_spec["value"]
                                map_type = self.get_type_declaration(map_spec)
                                map_name = self.gensym_decl(map_type, 'map')
                                with self.gensym_foreach_keyval(
                                        f'{iter_name}->second->{object_field}()'
                                ) as (map_key, map_value):
                                    match value_spec["kind"]:
                                        case "base" | "reference":
                                            with self.nest_name('value'):
                                                any_to_value = self.get_any_to_type_name(value_spec)
                                            map_expr = f'{any_to_value}(*{map_value})'
                                        case "array":
                                            array_spec = value_spec
                                            array_type = self.get_type_declaration(array_spec)
                                            array_name = self.gensym_decl(array_type, 'array')
                                            elem_spec = array_spec["element"]
                                            with self.nest_name('elem'):
                                                any_to_elem = self.get_any_to_type_name(elem_spec)
                                            with self.gensym_foreach(f'{map_value}->{array_field}()') as elem_name:
                                                self.gen_call(
                                                    f'{array_name}.push_back',
                                                    f'{any_to_elem}(*{elem_name})'
                                                )
                                            map_expr = f'std::move({array_name})'
                                        case "literal" | "or":
                                            with self.nest_name('value'):
                                                nested_name = self.nested_name()
                                                any_to_nested = self.get_any_to_type_name(nested_name)
                                                map_expr = f'{any_to_nested}(*{map_value})'
                                        case _:
                                            raise ValueError(
                                                f'Unsupported map value type ({value_spec["kind"]}): {value_spec}'
                                            )
                                    self.gen_call(f'{map_name}.emplace', map_key, map_expr)
                                field_expr = f'std::move({map_name})'
                            case "stringLiteral":
                                expected_value = field_type_spec["value"]
                                actual_value = self.gensym_ref('value', f'anyToString(*{iter_name}->second)')
                                with self.gen_if_ne(actual_value, f'"{expected_value}"'):
                                    self.gen_throw_invalid_params(
                                        f'("String value for "',
                                        f' "{struct_name}.{field_name}"',
                                        f' " must be \\"{expected_value}\\""',
                                        f' " but was: \\"{actual_value}\\"")'
                                    )
                                field_expr = actual_value
                            case _:
                                raise ValueError(
                                    'Unsupported property type ({kind}) for {struct}.{field}: {spec}'.format(
                                        kind=field_type_spec["kind"],
                                        struct=struct_name,
                                        field=field_name,
                                        spec=field_spec,
                                    )
                                )
                        if self.has_cycle(struct_symbol, field_spec["type"]):
                            field_type = self.get_type_declaration(field_spec["type"])
                            field_expr = f'std::make_unique<{field_type}>({field_expr})'
                        self.gen_assign(f'{value_name}.{field_name}', field_expr)
                    if not field_spec.get("optional", False):
                        with self.gen_else():
                            self.gen_throw_invalid_params(
                                f'"Missing required {struct_name} attribute: {field_name}"'
                            )
                    else:
                        self.inline(end='\n')
            self.newline()
            self.gen_return(value_name)

    def generate_struct_to_any_1(
            self,
            struct_symbol: LspSymbol,
            object_name: str,
            field_type_name: str,
            field_spec: LspSpec,
            field_value: str,
            is_optional: bool
    ) -> None:
        if is_optional:
            with self.gen_if(f'{field_value}.has_value()'):
                field_value = self.gensym_ref(
                    field_spec["name"],
                    f'{field_value}.value()'
                )
                if self.has_cycle(struct_symbol, field_spec["type"]):
                    field_value = f'*{field_value}'
                is_optional = False
                self.generate_struct_to_any_1(
                    struct_symbol,
                    object_name,
                    field_type_name,
                    field_spec,
                    field_value,
                    is_optional
                )
        else:
            field_name = field_spec["name"]
            field_type = field_spec["type"]
            match field_type["kind"]:
                case "base":
                    field_to_any = self.get_type_to_any_name(field_type)
                    field_expr = f'{field_to_any}({field_value})'
                    field_expr = f'std::make_unique<LSPAny>({field_expr})'
                case "reference":
                    match field_type["name"]:
                        case "LSPAny":
                            field_expr = f'copy({field_value})'
                        case "LSPObject" | "LSPArray":
                            any_name = self.gensym_decl('LSPAny', 'any')
                            self.gen_assign(any_name, f'copy({field_value})')
                            field_expr = f'std::move({any_name})'
                        case _:
                            field_to_any = self.get_type_to_any_name(field_type)
                            field_expr = f'{field_to_any}({field_value})'
                    field_expr = f'std::make_unique<LSPAny>({field_expr})'
                case "or" | "literal":
                    nested_name = self.nested_name()
                    nested_to_any = self.get_type_to_any_name(nested_name)
                    field_expr = f'{nested_to_any}({field_value})'
                    field_expr = f'std::make_unique<LSPAny>({field_expr})'
                case "array":
                    array_name = self.gensym_array(
                        'LSPArray', 'array',
                        f'{field_value}.size()'
                    )
                    elem_spec = field_type["element"]
                    with self.nest_name('elem'):
                        elem_to_any = self.get_type_to_any_name(elem_spec)
                    with self.gensym_foreach(field_value) as elem_name:
                        array_expr = elem_name
                        if self.has_cycle(struct_symbol, elem_spec):
                            array_expr = f'*{array_expr}'
                        match elem_spec["kind"]:
                            case "base" | "reference":
                                array_expr = f'{elem_to_any}({array_expr})'
                            case "or" | "literal":
                                with self.nest_name('elem') as nested_name:
                                    nested_to_any = self.get_type_to_any_name(nested_name)
                                    array_expr = f'{nested_to_any}({array_expr})'
                            case _:
                                raise ValueError(f'Unsupported array type ({elem_spec["kind"]}): {elem_spec}')
                        self.gen_call(
                            f'{array_name}.push_back',
                            self.move_as_any(array_expr, move_expr=False)
                        )
                    any_name = self.gen_as_any(array_name)
                    field_expr = f'std::move({any_name})'
                case "map":
                    map_spec = field_type
                    value_spec = map_spec["value"]
                    with self.nest_name('value'):
                        value_to_any = self.get_type_to_any_name(value_spec)
                    map_name = self.gensym_decl('LSPObject', 'map')
                    with self.gensym_foreach_keyval(field_value) as (key_name, value_name):
                        match value_spec["kind"]:
                            case "base" | "reference" | "or":
                                map_expr = f'{value_to_any}({value_name})'
                            case "array":
                                elem_spec = value_spec["element"]
                                with self.nest_name('elem'):
                                    elem_to_any = self.get_type_to_any_name(elem_spec)
                                array_name = self.gensym_decl('LSPArray', 'array')
                                with self.gensym_foreach(value_name) as elem_name:
                                    match elem_spec["kind"]:
                                        case "base":
                                            array_expr = f'{elem_to_any}({elem_name})'
                                        case "reference" | "or":
                                            array_expr = f'{elem_to_any}({elem_name})'
                                        case _:
                                            raise ValueError(
                                                f'Unsupported array type ({elem_spec["kind"]}): {elem_spec}'
                                            )
                                    self.gen_call(
                                        f'{array_name}.push_back',
                                        self.move_as_any(array_expr, move_expr=False)
                                    )
                                map_expr = array_name
                            case _:
                                raise ValueError(f'Unsupported map type ({value_spec["kind"]}): {value_spec}')
                        map_expr = self.move_as_any(map_expr, move_expr=False)
                        self.gen_call(f'{map_name}.emplace', key_name, map_expr)
                    field_expr = self.move_as_any(map_name)
                case "stringLiteral":
                    field_expr = f'stringToAny({field_value})'
                    field_expr = f'std::make_unique<LSPAny>({field_expr})'
                case _:
                    raise ValueError(f'Unsupported type ({field_type["kind"]}): {field_spec}')
            # if self.has_cycle(struct_symbol, field_type):
            #     field_expr = f'std::make_unique<{field_type_name}>({field_expr})'
            self.gen_call(f'{object_name}.emplace', f'"{field_name}"', field_expr)

    @gensym_context
    def generate_struct_to_any(self, struct_symbol: LspSymbol) -> None:
        struct_spec = struct_symbol.spec
        if "name" in struct_spec:
            struct_name = struct_spec["name"]
        else:
            struct_name = self.nested_name()
        field_types_and_specs = list(self.expand_fields(struct_spec))
        param_name = 'param'
        if len(field_types_and_specs) == 0:
            param_name = f'/*{param_name}*/'
        with self.gen_type_to_any(struct_name, f'const {struct_name} &{param_name}'):
            object_name = self.gensym_decl('LSPObject', 'object')
            self.newline()
            for field_type_name, field_spec in field_types_and_specs:
                field_name = field_spec["name"]
                with self.nested_names_as(deque([field_type_name, field_name])):
                    is_optional = field_spec.get("optional", False)
                    field_value = f'{param_name}.{field_name}'
                    self.generate_struct_to_any_1(
                        struct_symbol,
                        object_name,
                        field_type_name,
                        field_spec,
                        field_value,
                        is_optional
                    )
                    self.newline()
            self.gen_return_as_any(object_name)

    def generate_structure(self, struct_symbol: LspSymbol) -> None:
        with self.nested_name_as(struct_symbol.name):
            self.generate_any_to_struct(struct_symbol)
            self.generate_struct_to_any(struct_symbol)

    def generate_any_to_array_body(self, array_spec: LspSpec) -> None:
        elem_spec = array_spec["element"]
        array_field = rename_field("array")
        array_name = self.gensym_ref('array', f'any.{array_field}()')
        values_name = self.gensym_array(
            array_spec, 'values',
            f'{array_name}.size()'
        )
        with self.gensym_foreach(array_name) as elem_name:
            elem_expr = elem_name
            array_name = self.nested_name()
            match elem_spec["kind"]:
                case "base" | "reference" | "literal" | "or":
                    with self.nest_name('elem'):
                        any_to_elem = self.get_any_to_type_name(elem_spec)
                    elem_expr = f'{any_to_elem}(*{elem_expr})'
                case _:
                    raise ValueError(
                        f'Unsupported array type ({elem_spec["kind"]}): {elem_spec}'
                    )
            if self.has_cycle(array_name, elem_spec):
                elem_type = self.get_type_declaration(elem_spec)
                elem_expr = f'std::make_unique<{elem_type}>({elem_expr})'
            self.gen_call(f'{values_name}.push_back', elem_expr)
        self.gen_return(values_name)

    @gensym_context
    def generate_any_to_array(self, array_symbol: LspSymbol) -> None:
        array_name = array_symbol.name
        with self.gen_any_to_type(array_name, array_name):
            array_spec = array_symbol.spec
            self.generate_any_to_array_body(array_spec)

    @gensym_context
    def generate_any_to_alias(self, alias_symbol: LspSymbol) -> None:
        alias_name = alias_symbol.name
        alias_spec = alias_symbol.spec
        resolution = alias_symbol
        if alias_symbol.resolution is not None:
            resolution = alias_symbol.resolution
        with self.gen_any_to_type(alias_name, alias_name):
            type_spec = alias_spec["type"]
            match type_spec["kind"]:
                case "base":
                    type_enum = any_enum(resolution.spec['name'])
                    with self.gen_switch('any.type()'):
                        uinteger_field = rename_field('uinteger')
                        integer_field = rename_field('integer')
                        uinteger_type = rename_type('uinteger')
                        integer_type = rename_type('integer')
                        decimal_type = rename_type('decimal')
                        with self.gen_case('LSPAnyType', type_enum):
                            field_name = rename_field(resolution.normalized_name)
                            self.gen_return(f'any.{field_name}()')
                        match alias_name:
                            case "integer":
                                with self.gen_case('LSPAnyType', 'uinteger'):
                                    self.gen_return(
                                        f'static_cast<{integer_type}>(any.{uinteger_field}())'
                                    )
                            case "uinteger":
                                with self.gen_case('LSPAnyType', 'integer'):
                                    self.gen_return(
                                        f'static_cast<{uinteger_type}>(any.{integer_field}())'
                                    )
                            case "decimal":
                                with self.gen_case('LSPAnyType', 'integer'):
                                    self.gen_return(
                                        f'static_cast<{decimal_type}>(any.{integer_field}())'
                                    )
                                with self.gen_case('LSPAnyType', 'uinteger'):
                                    self.gen_return(
                                        f'static_cast<{decimal_type}>(any.{uinteger_field}())'
                                    )
                        with self.gen_default():
                            self.gen_throw_invalid_params(
                                f'("Cannot transform LSPAny of type LSPAnyType::" +',
                                f' LSPAnyTypeNames.at(any.type()) +',
                                f' " to type {rename_type(alias_name)}")'
                            )
                case "reference":
                    any_to_type = self.get_any_to_type_name(type_spec)
                    self.gen_return(f'{any_to_type}(any)')
                case "array":
                    return self.generate_any_to_array_body(type_spec)
                case _:
                    raise ValueError(
                        f'Unsupported alias type ({type_spec["kind"]}): {type_spec}'
                    )

    def generate_array_to_any_body(self, source_array: str, array_spec: LspSpec) -> None:
        elem_spec = array_spec["element"]
        target_array = \
            self.gensym_array('LSPArray', 'array', f'{source_array}.size()')
        with self.gensym_foreach(source_array) as elem_name:
            elem_expr = elem_name
            array_name = self.nested_name()
            if self.has_cycle(array_name, elem_spec):
                elem_expr = f'*{elem_expr}'
            with self.nest_name('elem'):
                elem_to_any = self.get_type_to_any_name(elem_spec)
            elem_expr = f'{elem_to_any}({elem_expr})'
            elem_expr = self.move_as_any(elem_expr, move_expr=False)
            self.gen_call(f'{target_array}.push_back', elem_expr)
        self.gen_return_as_any(target_array)

    @gensym_context
    def generate_array_to_any(self, array_symbol: LspSymbol) -> None:
        array_name = array_symbol.name
        array_type = rename_type(array_name)
        with self.gen_type_to_any(
                array_name,
                f'const {array_type} &array'
        ):
            self.generate_array_to_any_body('array', array_symbol.spec)

    @gensym_context
    def generate_alias_to_any(self, alias_symbol: LspSymbol) -> None:
        alias_name = alias_symbol.name
        alias_type = rename_type(alias_name)
        with self.gen_type_to_any(alias_name, f'const {alias_type} &alias'):
            alias_spec = alias_symbol.spec
            type_spec = alias_spec["type"]
            match type_spec["kind"]:
                case "base":
                    self.gen_return_as_any('alias', move=False)
                case "reference":
                    type_to_any = self.get_type_to_any_name(type_spec)
                    self.gen_return(f'{type_to_any}(alias)')
                case "array":
                    self.generate_array_to_any_body('alias', type_spec)
                case _:
                    raise ValueError(
                        f'Unsupported alias type ({type_spec["kind"]}): {type_spec}'
                    )

    def visit_alias_symbol(self, alias_symbol: LspSymbol) -> None:
        with self.nested_name_as(alias_symbol.name):
            self.generate_any_to_alias(alias_symbol)
            self.generate_alias_to_any(alias_symbol)

    @gensym_context
    def generate_base_to_any(self, base_symbol: LspSymbol) -> None:
        base_name = base_symbol.name
        base_type = rename_type(base_name)
        if base_symbol.normalized_name == "string":
            base_param = f'const {base_type} &param'
        else:
            base_param = f'{base_type} param'
        with self.gen_type_to_any(base_name, base_param):
            self.gen_return_as_any('param', move=False)

    @gensym_context
    def generate_any_to_base(self, base_symbol: LspSymbol) -> None:
        base_name = base_symbol.name
        base_type = rename_type(base_name)
        integer_type = rename_type('integer')
        uinteger_type = rename_type('uinteger')
        decimal_type = rename_type('decimal')
        integer_field = rename_field('integer')
        uinteger_field = rename_field('uinteger')
        with self.gen_any_to_type(base_name, base_type):
            with self.gen_switch('any.type()'):
                with self.gen_case('LSPAnyType', base_name):
                    field_name = rename_field(base_symbol.resolution.normalized_name)
                    self.gen_return(f'any.{field_name}()')
                match base_name:
                    case 'integer':
                        with self.gen_case('LSPAnyType', 'uinteger'):
                            self.gen_return(
                                f'static_cast<{integer_type}>(any.{uinteger_field}())'
                            )
                    case 'uinteger':
                        with self.gen_case('LSPAnyType', 'integer'):
                            self.gen_return(
                                f'static_cast<{uinteger_type}>(any.{integer_field}())'
                            )
                    case 'decimal':
                        with self.gen_case('LSPAnyType', 'integer'):
                            self.gen_return(
                                f'static_cast<{decimal_type}>(any.{integer_field}())'
                            )
                        with self.gen_case('LSPAnyType', 'uinteger'):
                            self.gen_return(
                                f'static_cast<{decimal_type}>(any.{uinteger_field}())'
                            )
                with self.gen_default():
                    self.gen_throw_invalid_params(
                        f'("Cannot transform LSPAny of type LSPAnyType::" +',
                        f' LSPAnyTypeNames.at(any.type()) +',
                        f' " to type {base_type}")'
                    )

    def visit_base_symbol(self, base_symbol: LspSymbol) -> None:
        with self.nested_name_as(base_symbol.name):
            self.generate_base_to_any(base_symbol)
            self.generate_any_to_base(base_symbol)

    def visit_array_symbol(self, array_symbol: LspSymbol) -> None:
        with self.nested_name_as(array_symbol.name):
            self.generate_array_to_any(array_symbol)
            self.generate_any_to_array(array_symbol)

    @gensym_context
    def generate_map_to_any(self, map_symbol: LspSymbol) -> None:
        map_name = map_symbol.name
        map_type = rename_type(map_name)
        map_param = f'const {map_type} &map'
        with self.gen_type_to_any(map_name, map_param):
            map_spec = map_symbol.spec
            value_spec = map_spec['value']
            with self.nest_name('value'):
                value_to_any = self.get_type_to_any_name(value_spec)
            object_name = self.gensym_decl('LSPObject', 'object')
            with self.gensym_foreach_keyval('map') as (map_key, map_value):
                value_expr = map_value
                if self.has_cycle(map_symbol, value_spec):
                    value_expr = f'*{value_expr}'
                value_expr = f'{value_to_any}({value_expr})'
                value_expr = f'std::make_unique<LSPAny>({value_expr})'
                self.gen_call(f'{object_name}.emplace', map_key, value_expr)
            self.gen_return_as_any(object_name)

    @gensym_context
    def generate_any_to_map(self, map_symbol: LspSymbol) -> None:
        map_name = map_symbol.name
        with self.gen_any_to_type(map_name, map_name):
            map_spec = map_symbol.spec
            value_spec = map_spec['value']
            with self.nest_name('value'):
                any_to_value = self.get_any_to_type_name(value_spec)
            map_name = self.gensym_decl(map_spec, 'map')
            object_field = rename_field('object')
            with self.gensym_foreach_keyval(f'any.{object_field}()') as (map_key, map_value):
                value_expr = f'{any_to_value}(*{map_value})'
                if self.has_cycle(map_symbol, value_spec):
                    value_type = self.get_type_declaration(value_spec)
                    value_expr = f'std::make_unique<{value_type}>({value_expr})'
                self.gen_call(f'{map_name}.emplace', map_key, value_expr)
            self.gen_return(map_name)

    def visit_map_symbol(self, map_symbol: LspSymbol) -> None:
        with self.nested_name_as(map_symbol.name):
            self.generate_map_to_any(map_symbol)
            self.generate_any_to_map(map_symbol)

    @gensym_context
    def generate_copy_any(self) -> None:
        with self.gen_fn(
                'LspTransformer::copy',
                'LSPAny',
                params=[
                    'const LSPAny &any',
                ],
                specs="const"
        ):
            copy_name = self.gensym_decl('LSPAny', 'any')
            with self.gen_switch('any.type()'):
                with self.gen_case("LSPAnyType", "object"):
                    object_field = rename_field("LSPObject")
                    self.gen_assign(copy_name, f'copy(any.{object_field}())')
                    self.gen_break()
                with self.gen_case('LSPAnyType', 'array'):
                    array_field = rename_field("LSPArray")
                    self.gen_assign(copy_name, f'copy(any.{array_field}())')
                    self.gen_break()
                with self.gen_case('LSPAnyType', 'string'):
                    string_field = rename_field("string")
                    self.gen_assign(copy_name, f'any.{string_field}()')
                    self.gen_break()
                with self.gen_case('LSPAnyType', 'integer'):
                    integer_field = rename_field("integer")
                    self.gen_assign(copy_name, f'any.{integer_field}()')
                    self.gen_break()
                with self.gen_case('LSPAnyType', 'uinteger'):
                    uinteger_field = rename_field("uinteger")
                    self.gen_assign(copy_name, f'any.{uinteger_field}()')
                    self.gen_break()
                with self.gen_case('LSPAnyType', 'decimal'):
                    decimal_field = rename_field("decimal")
                    self.gen_assign(copy_name, f'any.{decimal_field}()')
                    self.gen_break()
                with self.gen_case('LSPAnyType', f'boolean'):
                    boolean_field = rename_field("boolean")
                    self.gen_assign(copy_name, f'any.{boolean_field}()')
                    self.gen_break()
                with self.gen_case('LSPAnyType', 'null'):
                    null_field = rename_field("null")
                    self.gen_assign(copy_name, f'any.{null_field}()')
                    self.gen_break()
                with self.gen_case('LSPAnyType', 'Uninitialized'):
                    self.gen_throw(
                        'std::invalid_argument',
                        '"LSPAny has not been initialized."'
                    )
            self.gen_return(copy_name)
        self.newline()

    @gensym_context
    def generate_copy_any_pointer(self) -> None:
        with self.gen_fn(
                'LspTransformer::copy',
                'std::unique_ptr<LSPAny>',
                params=[
                    'const std::unique_ptr<LSPAny> &any',
                ],
                specs="const"
        ):
            any_name = self.gensym_init('any', 'std::make_unique<LSPAny>()')
            self.gen_assign(f'(*{any_name})', 'copy(*any)')
            self.gen_return(any_name)
        self.newline()

    @gensym_context
    def generate_copy_object(self) -> None:
        with self.gen_fn(
                'LspTransformer::copy',
                'LSPObject',
                params=[
                    'const LSPObject &object',
                ],
                specs='const'
        ):
            copy_name = self.gensym_decl('LSPObject', 'copy')
            with self.gensym_foreach_keyval('object') as (key_name, value_name):
                self.gen_call(
                    f'{copy_name}.emplace',
                    key_name,
                    f'copy({value_name})'
                )
            self.gen_return(copy_name)
        self.newline()

    @gensym_context
    def generate_copy_array(self) -> None:
        with self.gen_fn(
                'LspTransformer::copy',
                'LSPArray',
                params=[
                    'const LSPArray &array',
                ],
                specs='const'
        ):
            copy_name = self.gensym_array('LSPArray', 'copy', 'array.size()')
            with self.gensym_foreach('array') as elem_name:
                self.gen_call(f'{copy_name}.push_back', f'copy({elem_name})')
            self.gen_return(copy_name)
        self.newline()

    def generate_copy_methods(self) -> None:
        self.generate_copy_any()
        self.generate_copy_any_pointer()
        self.generate_copy_object()
        self.generate_copy_array()

    def get_as_message_params_type(self, type_spec: LspSpec) -> str:
        match type_spec["kind"]:
            case "reference":
                return type_spec["name"]
            case _:
                raise ValueError(f'Unsupported request parameter type: {type_spec}')

    @gensym_context
    def generate_as_incoming_params(
        self,
        message_name: str,
        message_symbol: LspSymbol
    ) -> None:
        message_spec = message_symbol.spec
        params_spec = message_spec.get("params", None)
        if params_spec is not None:
            method_name = method_to_camel_case(message_spec["method"])
            type_name = f'{method_name}Params'
            as_message_params = f'as{type_name}'
            with self.gen_fn(
                    f'LspTransformer::{as_message_params}',
                    self.get_as_message_params_type(params_spec),
                    params=[
                        f'const MessageParams &{message_name}'
                    ],
                    specs='const'
            ):
                object_enum = rename_enum('object')
                object_field = rename_field('LSPObject')
                with self.gen_if_ne(f'{message_name}.type()', f'MessageParamsType::{object_enum}'):
                    self.gen_throw_invalid_params(
                        f'("Message parameter type must be MessageParamsType::{object_enum} "',
                        f' "for method=\\"{message_spec["method"]}\\" but received "',
                        f' "MessageParamsType::" + MessageParamsTypeNames.at({message_name}.type()))'
                    )
                self.newline()
                match params_spec["kind"]:
                    case "reference":
                        params_type = self.get_type_declaration(params_spec)
                        params_name = self.gensym_decl(params_type, 'params')
                        field_types_and_specs = list(self.expand_fields(params_spec))
                        if len(field_types_and_specs) > 0:
                            object_name = self.gensym_ref('object', f'{message_name}.{object_field}()')
                            iter_name = self.gensym_decl('LSPObject::const_iterator', 'iter')
                            self.newline()
                            for field_type_name, field_spec in field_types_and_specs:
                                field_name = field_spec["name"]
                                with self.nested_names_as([field_type_name, field_name]):
                                    field_type = field_spec["type"]
                                    self.gen_assign(f'{iter_name}', f'{object_name}.find("{field_name}")')
                                    with self.gen_if_ne(iter_name, f'{object_name}.end()', end=''):
                                        match field_type["kind"]:
                                            case "base":
                                                base_spec = field_type
                                                any_to_base = self.get_any_to_type_name(base_spec)
                                                params_expr = f'{any_to_base}(*{iter_name}->second)'
                                            case "reference":
                                                match field_type["name"]:
                                                    case "LSPAny":
                                                        params_expr = f'copy(*{iter_name}->second)'
                                                    case "LSPObject" | "LSPArray":
                                                        params_expr = f'copy(*{iter_name}->second)'
                                                    case _:
                                                        type_spec = field_type
                                                        any_to_type = self.get_any_to_type_name(type_spec)
                                                        params_expr = f'{any_to_type}(*{iter_name}->second)'
                                            case "array":
                                                array_spec = field_type
                                                elem_spec = array_spec["element"]
                                                array_field = rename_field('LSPArray')
                                                array_name = self.gensym_ref('array', f'{iter_name}->second->{array_field}()')
                                                values_name = self.gensym_decl(array_spec, 'values')
                                                self.write(f'{values_name}.reserve({array_name}.size());')
                                                with self.gensym_foreach(array_name) as elem_name:
                                                    match elem_spec["name"]:
                                                        case "LSPAny":
                                                            elem_expr = f'copy(*{elem_name})'
                                                        case "LSPObject":
                                                            elem_expr = f'copy({elem_name}->{object_field}())'
                                                        case "LSPArray":
                                                            elem_expr = f'copy({elem_name}->{array_field}())'
                                                        case _:
                                                            with self.nest_name('elem'):
                                                                any_to_elem = self.get_any_to_type_name(elem_spec)
                                                            elem_expr = f'{any_to_elem}(*{elem_name})'
                                                    self.gen_call(f'{values_name}.push_back', elem_expr)
                                                params_expr = f'std::move({values_name})'
                                            case "or" | "literal":
                                                nested_name = self.nested_name([field_type_name, field_name])
                                                any_to_nested = self.get_any_to_type_name(nested_name)
                                                params_expr = f'{any_to_nested}(*{iter_name}->second)'
                                            case _:
                                                raise ValueError(
                                                    f'Unsupported property kind ({field_type["kind"]}): {field_type}'
                                                )
                                        self.gen_assign(f'{params_name}.{field_name}', params_expr)
                                    if not field_spec.get("optional", False):
                                        with self.gen_else():
                                            self.gen_throw_invalid_params(
                                                f'"Missing required {params_spec["name"]} attribute: {field_name}"'
                                            )
                                    else:
                                        self.inline(end='\n')
                                    self.newline()
                        self.gen_return(params_name)
                    case _:
                        raise ValueError(
                            f'Unsupported message parameter type ({params_spec["name"]}): {params_spec}'
                        )
            self.newline()

    def generate_result_union_array_union_to_any(
            self,
            result_name: str,
            array_name: str,
            union_spec: LspSpec,
            index: int
    ) -> None:
        union_name = self.nested_name([result_name, str(index)])
        field_name = rename_field(union_name)
        field_value = f'result.{field_name}()'
        self.gen_call(
            f'{array_name}.reserve',
            f'{field_value}.size()'
        )
        union_enum = f'{union_name}Type'
        with self.gensym_foreach(field_value) as elem_name:
            with self.gen_switch(f'{elem_name}.type()'):
                for union_item in union_spec["items"]:
                    with self.gen_case(union_enum, union_item):
                        match union_item["kind"]:
                            case "base" | "reference":
                                item_field = self.name_field(union_item)
                                item_name = union_item["name"]
                                match item_name:
                                    case "LSPAny" | "LSPObject" | "LSPArray":
                                        item_field = rename_field(item_name)
                                        elem_expr = f'copy({elem_name}.{item_field}())'
                                    case _:
                                        item_to_any = self.get_type_to_any_name(union_item)
                                        elem_expr = f'{item_to_any}({elem_name}.{item_field}())'
                            case _:
                                raise ValueError(
                                    f'Unsupported item kind ({union_item["kind"]}): {union_item}'
                                )
                        self.gen_call(f'{array_name}.push_back', elem_expr)
                        self.gen_break()
                with self.gen_case(union_enum, 'Uninitialized'):
                    self.gen_throw(
                        'LSP_EXCEPTION',
                        'ErrorCodes::InternalError',
                        f'"{result_name} has not been initialized!"'
                    )

    def generate_result_union_array_to_any(
            self,
            result_name: str,
            array_spec: LspSpec,
            index: int
    ) -> None:
        field_name = self.name_field(array_spec)
        elem_spec = array_spec["element"]
        array_name = self.gensym_decl('LSPArray', 'array')
        match elem_spec["kind"]:
            case "base" | "reference":
                field_value = f'result.{field_name}()'
                self.gen_call(
                    f'{array_name}.reserve',
                    f'{field_value}.size()'
                )
                with self.gensym_foreach(field_value) as elem_name:
                    with self.nest_name('elem'):
                        elem_to_any = self.get_type_to_any_name(elem_spec)
                    elem_expr = f'{elem_to_any}({elem_name})'
                    self.gen_call(
                        f'{array_name}.push_back',
                        self.move_as_any(elem_expr, move_expr=False)
                    )
            case "or":
                self.generate_result_union_array_union_to_any(
                    result_name,
                    array_name,
                    elem_spec,
                    index
                )
            case _:
                raise ValueError(f'Unsupported array type ({elem_spec["kind"]}): {elem_spec}')
        self.gen_return_as_any(array_name)

    def generate_result_union_reference_to_any(
            self,
            reference_spec: LspSpec,
            field_expr: str
    ) -> None:
        reference_to_any = self.get_type_to_any_name(reference_spec)
        reference_name = reference_spec["name"]
        match reference_name:
            case "LSPAny" | "LSPObject" | "LSPArray":
                self.gen_return(f'copy({field_expr})')
            case _:
                reference_symbol = self.resolve(reference_name)
                match reference_symbol.kind:
                    case LspSymbolKind.BASE:
                        if reference_symbol.normalized_name == "string":
                            field_expr = f'*{field_expr}'
                    case LspSymbolKind.STRUCTURE | \
                         LspSymbolKind.INNER | \
                         LspSymbolKind.UNION:
                        field_expr = f'*{field_expr}'
                self.gen_return(f'{reference_to_any}({field_expr})')

    def generate_result_union_to_any(
            self,
            result_type: str,
            result_name: str,
            symbol_spec: LspSpec
    ) -> None:
        with self.gen_switch('result.type()'):
            for index, item_spec in enumerate(symbol_spec["items"]):
                with self.gen_case(result_type, item_spec):
                    match item_spec["kind"]:
                        case "base":
                            field_name = self.name_field(item_spec)
                            item_to_any = self.get_type_to_any_name(item_spec)
                            self.gen_return(f'{item_to_any}(result.{field_name}())')
                        case "reference":
                            reference_spec = item_spec
                            reference_name = reference_spec["name"]
                            field_name = rename_field(reference_name)
                            field_expr = f'result.{field_name}()'
                            self.generate_result_union_reference_to_any(
                                reference_spec,
                                field_expr
                            )
                        case "array":
                            with self.nest_name('elem'):
                                self.generate_result_union_array_to_any(
                                    result_name,
                                    item_spec,
                                    index
                                )
                        case _:
                            raise ValueError(f'Unsupported union type ({item_spec["kind"]}): {item_spec}')
            with self.gen_case(result_type, 'Uninitialized'):
                self.gen_throw(
                    'LSP_EXCEPTION',
                    'ErrorCodes::InternalError',
                    f'"{result_name} has not been initialized!"'
                )
        self.write('throw LSP_EXCEPTION(')
        with self.indent():
            self.write('ErrorCodes::InternalError,')
            self.write('"Should be unreachable."')
        self.write(');')

    def generate_result_array_to_any(self, array_spec: LspSpec) -> None:
        elem_spec = array_spec["element"]
        array_name = self.gensym_array('LSPArray', 'array', 'result.size()')
        with self.gensym_foreach('result') as elem_name:
            with self.nest_name('elem'):
                elem_to_any = self.get_type_to_any_name(elem_spec)
            elem_expr = f'{elem_to_any}({elem_name})'
            self.gen_call(
                f'{array_name}.push_back',
                self.move_as_any(elem_expr, move_expr=False)
            )
        self.gen_return_as_any(array_name)

    @gensym_context
    def generate_result_to_any(self, request_spec: LspSpec) -> None:
        request_method = request_spec["method"]
        request_name = method_to_camel_case(request_method)
        result_name = f'{request_name}Result'
        with self.nested_names_as(deque([result_name])):
            result_type = f'{result_name}Type'
            symbol = self.resolve(result_name)
            symbol_spec = symbol.spec
            with self.gen_type_to_any(result_name, f'const {result_name} &result'):
                match symbol.kind:
                    case LspSymbolKind.UNION:
                        self.generate_result_union_to_any(
                            result_type,
                            result_name,
                            symbol_spec
                        )
                    case LspSymbolKind.BASE | LspSymbolKind.REFERENCE:
                        type_to_any = self.get_type_to_any_name(symbol)
                        self.gen_return(f'{type_to_any}(result)')
                    case LspSymbolKind.ARRAY:
                        self.generate_result_array_to_any(symbol_spec)
                    case _:
                        raise ValueError(f'Unsupported symbol kind ({symbol.kind}): {symbol}')

    def gen_struct_as_outgoing_params(
            self,
            object_name: str,
            field_name: str,
            field_spec: LspSpec,
            field_value: str,
            is_optional: bool
    ) -> None:
        if is_optional:
            with self.gen_if(f'{field_value}.has_value()'):
                is_optional = False
                field_value = f'{field_value}.value()'
                self.gen_struct_as_outgoing_params(
                    object_name,
                    field_name,
                    field_spec,
                    field_value,
                    is_optional
                )
        else:
            field_type = field_spec["type"]
            match field_type["kind"]:
                case 'base' | 'reference' | 'or' | 'literal':
                    type_to_any = self.get_type_to_any_name(field_type)
                    field_expr = f'{type_to_any}({field_value})'
                case 'array':
                    array_spec = field_type
                    elem_spec: LspSpec = array_spec['element']
                    with self.nest_name('elem'):
                        elem_to_any = self.get_type_to_any_name(elem_spec)
                    array_name = self.gensym_array(
                        'LSPArray', 'array',
                        f'{field_value}.size()'
                    )
                    with self.gensym_foreach(field_value) as elem_name:
                        elem_expr = f'{elem_to_any}({elem_name})'
                        elem_expr = self.move_as_any(elem_expr, move_expr=False)
                        self.gen_call(f'{array_name}.push_back', elem_expr)
                    field_expr = array_name
                case _:
                    raise ValueError(f'Unsupported property kind ({field_type["kind"]}): {field_type}')
            self.gen_call(
                f'{object_name}.emplace',
                f'"{field_name}"',
                self.move_as_any(field_expr, move_expr=False)
            )

    @gensym_context
    def generate_as_outgoing_params(
        self,
        params_name: str,
        message_symbol: LspSymbol
    ) -> None:
        message_spec = message_symbol.spec
        params_spec = message_spec.get("params", None)
        if params_spec is not None:
            params_type = params_spec["name"]
            with self.gen_fn(
                    'LspTransformer::asMessageParams',
                    'MessageParams',
                    params=[
                        f'const {params_type} &{params_name}',
                    ],
                    specs='const'
            ):
                outgoing_name = self.gensym_decl('MessageParams', 'messageParams')
                match params_spec["kind"]:
                    case "reference":
                        type_name = params_spec["name"]
                        match type_name:
                            case "LSPAny":
                                with self.gen_switch(f'{params_name}.type()'):
                                    with self.gen_case('LSPAnyType', 'object'):
                                        object_field = rename_field("LSPObject")
                                        object_name = self.gensym_ref('object', f'{params_name}.{object_field}()')
                                        self.gen_assign(f'{outgoing_name}', f'copy({object_name})')
                                        self.gen_break()
                                    with self.gen_case('LSPAnyType', 'array'):
                                        array_field = rename_field("LSPArray")
                                        array_name = self.gensym_ref('array', f'{params_name}.{array_field}()')
                                        self.gen_assign(f'{outgoing_name}', f'copy({array_name})')
                                        self.gen_break()
                                    with self.gen_default():
                                        self.gen_throw_invalid_params(
                                            f'("Invalid LSPAny type for {params_name}: LSPAnyType::" +',
                                            f' LSPAnyTypeNames.at({params_name}.type()))'
                                        )
                                self.newline()
                            case _:
                                symbol = self.resolve(type_name)
                                object_name = self.gensym_decl('LSPObject', 'object')
                                self.newline()
                                for field_type_name, field_spec in self.expand_fields(symbol.spec):
                                    field_name = field_spec['name']
                                    with self.nested_names_as(deque([field_type_name, field_name])):
                                        is_optional = field_spec.get('optional', False)
                                        field_value = f'{params_name}.{field_name}'
                                        self.gen_struct_as_outgoing_params(
                                            object_name,
                                            field_name,
                                            field_spec,
                                            field_value,
                                            is_optional
                                        )
                                        self.newline()
                                self.gen_assign(
                                    f'{outgoing_name}',
                                    f'std::move({object_name})'
                                )
                    case _:
                        raise ValueError(
                            f'Unsupported parameter type ({params_spec["kind"]}): {params_spec}'
                        )
                self.gen_return(outgoing_name)
            self.newline()

    def visit_request_symbol(self, request_symbol: LspSymbol) -> None:
        request_spec = request_symbol.spec
        match request_spec["messageDirection"]:
            case "clientToServer":
                self.generate_as_incoming_params("requestParams", request_symbol)
            case "serverToClient":
                self.generate_as_outgoing_params("requestParams", request_symbol)
            case "both":
                self.generate_as_incoming_params("requestParams", request_symbol)
                self.generate_as_outgoing_params("requestParams", request_symbol)
            case _:
                raise ValueError(
                    f'Unsupported messageDirection: {request_spec["messageDirection"]}'
                )

    def visit_notification_symbol(self, notification_symbol: LspSymbol) -> None:
        notification_spec = notification_symbol.spec
        match notification_spec["messageDirection"]:
            case "clientToServer":
                self.generate_as_incoming_params("notificationParams", notification_symbol)
            case "serverToClient":
                self.generate_as_outgoing_params("notificationParams", notification_symbol)
            case "both":
                self.generate_as_incoming_params("notificationParams", notification_symbol)
                self.generate_as_outgoing_params("notificationParams", notification_symbol)
            case _:
                raise ValueError(
                    f'Unsupported messageDirection: {notification_spec["messageDirection"]}'
                )

    def generate_code(self) -> None:
        print(f'Generating: {self.file_path}')
        self.generate_disclaimer()
        self.gen_include('stdexcept')
        self.newline()
        self.gen_include('server/lsp_exception.h')
        self.gen_include('server/lsp_specification.h')
        self.gen_include('server/lsp_transformer.h')
        self.newline()
        with self.gen_namespace(self.namespace):
            self.newline()
            super().generate_code()
            self.generate_copy_methods()



import json
from pathlib import Path

root = Path(__file__).parent

with open(root / "_static" / "languages.json", "r", encoding="utf-8") as fd:
    all_languages = json.load(fd)

if "language" not in globals():
    language = "en"

project = "LFortran"
author = f"{project} contributors"
copyright = f"2022 {author}"

extensions = [
    "myst_parser",
    "nbsphinx",
    "sphinx_copybutton",
]
myst_enable_extensions = [
    "dollarmath",
]

html_theme = "sphinx_material"
html_title = project
html_logo = "_static/img/lfortran-logo.svg"
html_favicon = html_logo
html_css_files = ["css/custom.css", "css/ansi.css"]
html_static_path = ["_static"]
html_show_sourcelink = False

html_theme_options = {
    "nav_title": f"{project} Documentation",
    "base_url": f"https://docs.lfortran.org/{language}/",
    "repo_name": project,
    "color_primary": "amber",
    "color_accent": "orange",
    "repo_url": "https://github.com/lfortran/lfortran",
    "nav_links": [
        dict(href="installation", internal=True, title="Getting started"),
        dict(href="progress", internal=True, title="LFortran Development Status"),
        dict(href="design", internal=True, title="Developers's Guide"),
        dict(href="intrinsics/array", internal=True, title="Intrinsics"),
    ],
    "version_info": {
        lang: f"/../{code}/"
        for lang, code in all_languages.items()
    },
    "globaltoc_depth": 2,
}

html_sidebars = {
    "**": ["logo-text.html", "globaltoc.html", "localtoc.html", "searchbox.html"]
}

templates_path = ["_templates"]
locale_dirs = ["../locale"]
gettext_compact = "docs"
